from mindsdb.api.http.namespaces.configs.predictors import ns_conf
from mindsdb.api.http.namespaces.entitites.column_metadata import column_metadata

from flask_restx import fields

confusion_matrix_data = ns_conf.model('ConfusionMatrixData', {
    'matrix': fields.List(fields.List(fields.Integer, required=True)),
    'predicted': fields.List(fields.String, required=False, description='Predicted values'),
    'real': fields.List(fields.String, required=False, description='Real values'),
})



import pg8000

from mindsdb_native.libs.constants.mindsdb import DATA_SUBTYPES
from mindsdb.integrations.base import Integration


class PostgreSQL(Integration):
    def _to_postgres_table(self, stats, predicted_cols):
        subtype_map = {
            DATA_SUBTYPES.INT: ' int8',
            DATA_SUBTYPES.FLOAT: 'float8',
            DATA_SUBTYPES.BINARY: 'bool',
            DATA_SUBTYPES.DATE: 'date',
            DATA_SUBTYPES.TIMESTAMP: 'timestamp',
            DATA_SUBTYPES.SINGLE: 'text',
            DATA_SUBTYPES.MULTIPLE: 'text',
            DATA_SUBTYPES.TAGS: 'text',
            DATA_SUBTYPES.IMAGE: 'text',
            DATA_SUBTYPES.VIDEO: 'text',
            DATA_SUBTYPES.AUDIO: 'text',
            DATA_SUBTYPES.SHORT: 'text',
            DATA_SUBTYPES.RICH: 'text',
            DATA_SUBTYPES.ARRAY: 'text'
        }

        column_declaration = []
        for name in stats['columns']:
            try:
                col_subtype = stats[name]['typing']['data_subtype']
                new_type = subtype_map[col_subtype]
                column_declaration.append(f' "{name}" {new_type} ')
                if name in predicted_cols:
                    column_declaration.append(f' "{name}_original" {new_type} ')
            except Exception:
                print(f'Error: cant convert type {col_subtype} of column {name} to Postgres type')

        return column_declaration

    def _escape_table_name(self, name):
        return '"' + name.replace('"', '""') + '"'

    def _query(self, query):
        con = pg8000.connect(
            database=self.config['integrations'][self.name].get('database', 'postgres'),
            user=self.config['integrations'][self.name]['user'],
            password=self.config['integrations'][self.name]['password'],
            host=self.config['integrations'][self.name]['host'],
            port=self.config['integrations'][self.name]['port']
        )

        cur = con.cursor()
        res = True
        cur.execute(query)

        try:
            rows = cur.fetchall()
            keys = [k[0] if isinstance(k[0], str) else k[0].decode('ascii') for k in cur.description]
            res = [dict(zip(keys, row)) for row in rows]
        except Exception:
            pass

        con.commit()
        con.close()

        return res

    def setup(self):
        user = f"{self.config['api']['mysql']['user']}_{self.name}"
        password = self.config['api']['mysql']['password']
        host = self.config['api']['mysql']['host']
        port = self.config['api']['mysql']['port']

        try:
            self._query('''
                DO $$
                begin
                    if not exists (SELECT 1 FROM pg_extension where extname = 'mysql_fdw') then
                        CREATE EXTENSION mysql_fdw;
                    end if;
                END
                $$;
            ''')
        except Exception:
            print('Error: cant find or activate mysql_fdw extension for PostgreSQL.')

        self._query(f'DROP SCHEMA IF EXISTS {self.mindsdb_database} CASCADE')

        self._query(f"DROP USER MAPPING IF EXISTS FOR {self.config['integrations'][self.name]['user']} SERVER server_{self.mindsdb_database}")

        self._query(f'DROP SERVER IF EXISTS server_{self.mindsdb_database} CASCADE')

        self._query(f'''
            CREATE SERVER server_{self.mindsdb_database}
                FOREIGN DATA WRAPPER mysql_fdw
                OPTIONS (host '{host}', port '{port}');
        ''')

        self._query(f'''
           CREATE USER MAPPING FOR {self.config['integrations'][self.name]['user']}
                SERVER server_{self.mindsdb_database}
                OPTIONS (username '{user}', password '{password}');
        ''')

        self._query(f'CREATE SCHEMA {self.mindsdb_database}')

        q = f"""
            CREATE FOREIGN TABLE IF NOT EXISTS {self.mindsdb_database}.predictors (
                name text,
                status text,
                accuracy text,
                predict text,
                select_data_query text,
                external_datasource text,
                training_options text
            )
            SERVER server_{self.mindsdb_database}
            OPTIONS (dbname 'mindsdb', table_name 'predictors');
        """
        self._query(q)

        q = f"""
            CREATE FOREIGN TABLE IF NOT EXISTS {self.mindsdb_database}.commands (
                command text
            ) SERVER server_{self.mindsdb_database}
            OPTIONS (dbname 'mindsdb', table_name 'commands');
        """
        self._query(q)

    def register_predictors(self, model_data_arr):
        for model_meta in model_data_arr:
            name = model_meta['name']
            stats = model_meta['data_analysis_v2']
            columns_sql = ','.join(self._to_postgres_table(stats, model_meta['predict']))
            columns_sql += ',"select_data_query" text'
            columns_sql += ',"external_datasource" text'
            for col in model_meta['predict']:
                columns_sql += f',"{col}_confidence" float8'
                if model_meta['data_analysis_v2'][col]['typing']['data_type'] == 'Numeric':
                    columns_sql += f',"{col}_min" float8'
                    columns_sql += f',"{col}_max" float8'
                columns_sql += f',"{col}_explain" text'

            q = f"""
                CREATE FOREIGN TABLE {self.mindsdb_database}.{self._escape_table_name(name)} (
                    {columns_sql}
                ) SERVER server_{self.mindsdb_database}
                OPTIONS (dbname 'mindsdb', table_name '{name}');
            """
            self._query(q)

    def unregister_predictor(self, name):
        q = f"""
            DROP FOREIGN TABLE IF EXISTS {self.mindsdb_database}.{self._escape_table_name(name)};
        """
        self._query(q)

    def check_connection(self):
        try:
            con = pg8000.connect(
                database=self.config['integrations'][self.name].get('database', 'postgres'),
                user=self.config['integrations'][self.name]['user'],
                password=self.config['integrations'][self.name]['password'],
                host=self.config['integrations'][self.name]['host'],
                port=self.config['integrations'][self.name]['port']
            )
            con.run('select 1;')
            con.close()
            connected = True
        except Exception:
            connected = False
        return connected



"""
*******************************************************
 * Copyright (C) 2017 MindsDB Inc. <copyright@mindsdb.com>
 *
 * This file is part of MindsDB Server.
 *
 * MindsDB Server can not be copied and/or distributed without the express
 * permission of MindsDB Inc
 *******************************************************
"""

import struct
import math

from mindsdb.api.mysql.mysql_proxy.data_types.mysql_packet import Packet
from mindsdb.api.mysql.mysql_proxy.data_types.mysql_datum import Datum
from mindsdb.api.mysql.mysql_proxy.libs.constants.mysql import COMMANDS, getConstName, TYPES

from mindsdb.api.mysql.mysql_proxy.classes.sql_statement_parser import SQL_PARAMETER


class CommandPacket(Packet):
    '''
    Implementation based on description:
    https://mariadb.com/kb/en/library/1-connecting-connecting/#initial-handshake-packet
    '''

    def _read_byte(self, buffer):
        b = buffer[:1]
        buffer = buffer[1:]
        b = struct.unpack('<B', b)[0]
        return b, buffer

    def setup(self, length=0, count_header=1, body=''):
        if length == 0:
            return

        # self.salt=self.session.salt

        self._length = length
        self._seq = count_header
        self._body = body

        self.type = Datum('int<1>')
        buffer = body
        buffer = self.type.setFromBuff(buffer)

        if self.type.value in (COMMANDS.COM_QUERY, COMMANDS.COM_STMT_PREPARE):
            self.sql = Datum('str<EOF>')
            buffer = self.sql.setFromBuff(buffer)
        elif self.type.value == COMMANDS.COM_STMT_EXECUTE:
            # https://mariadb.com/kb/en/com_stmt_execute/
            self.stmt_id = Datum('int<4>')
            buffer = self.stmt_id.setFromBuff(buffer)
            self.flags = Datum('int<1>')
            buffer = self.flags.setFromBuff(buffer)
            self.iteration_count = Datum('int<4>')
            buffer = self.iteration_count.setFromBuff(buffer)

            self.parameters = []

            prepared_stmt = self.session.prepared_stmts[self.stmt_id.value]

            if prepared_stmt['type'] in ['insert', 'delete']:
                if prepared_stmt['type'] == 'insert':
                    num_params = prepared_stmt['statement'].struct['values'].count(SQL_PARAMETER)
                elif prepared_stmt['type'] == 'delete':
                    num_params = prepared_stmt['statement'].sql.count('?')

                if num_params > 0:
                    # read null-map
                    null_bytes = math.floor((num_params + 7) / 8)
                    nulls = []
                    for i in range(null_bytes):
                        b, buffer = self._read_byte(buffer)
                        for i in range(8):
                            nulls.append(((1 << i) & b) != 0)

                    # read send-type byte
                    b, buffer = self._read_byte(buffer)

                    types = []
                    if b == 1:
                        # read types
                        for i in range(num_params):
                            t, buffer = self._read_byte(buffer)
                            s, buffer = self._read_byte(buffer)
                            types.append(dict(
                                type=t,
                                signed=s
                            ))

                    self.parameters = []
                    for i in range(num_params):
                        if nulls[i]:
                            self.parameters.append(None)
                            continue
                        if types[i]['type'] in (TYPES.MYSQL_TYPE_VAR_STRING, TYPES.MYSQL_TYPE_STRING):
                            x = Datum('string<lenenc>')
                            buffer = x.setFromBuff(buffer)
                            self.parameters.append(x.value.decode())
                        else:
                            # NOTE at this moment all sends as strings and it works
                            raise Exception(f"Unsupported type {types[i]['type']}")

        elif self.type.value == COMMANDS.COM_STMT_CLOSE:
            self.stmt_id = Datum('int<4>')
            buffer = self.stmt_id.setFromBuff(buffer)
        elif self.type.value == COMMANDS.COM_STMT_FETCH:
            self.stmt_id = Datum('int<4>')
            buffer = self.stmt_id.setFromBuff(buffer)
            self.limit = Datum('int<4>')
            buffer = self.limit.setFromBuff(buffer)
        elif self.type.value == COMMANDS.COM_INIT_DB:
            self.database = Datum('str<EOF>')
            buffer = self.database.setFromBuff(buffer)
        else:
            self.data = Datum('str<EOF>')
            buffer = self.data.setFromBuff(buffer)

    def __str__(self):
        return str({
            'header': {'length': self.length, 'seq': self.seq},
            'type': getConstName(COMMANDS, self.type.value),
            'vars': self.__dict__
        })



from mindsdb.api.mysql.mysql_proxy.datahub.datanodes.datanode import DataNode

class InformationSchema(DataNode):
    type = 'INFORMATION_SCHEMA'

    index = {}

    information_schema = {
        'SCHEMATA': ['schema_name', 'default_character_set_name', 'default_collation_name'],
        'TABLES': ['table_schema', 'table_name', 'table_type'],
        'COLUMNS': ['table_schema', 'table_name', 'ordinal_position'],
        'EVENTS': ['event_schema', 'event_name'],
        'ROUTINES': ['routine_schema', 'specific_name', 'routine_type'],
        'TRIGGERS': ['trigger_schema', 'trigger_name']
    }

    def __init__(self, dsObject=None):
        if isinstance(dsObject, dict):
            self.add(dsObject)

    def __getitem__(self, key):
        return self.get(key)

    def add(self, dsObject):
        for key, val in dsObject.items():
            self.index[key.upper()] = val

    def get(self, name):
        # INFORMATION_SCHEMA.SCHEMATA
        if name.upper() == 'INFORMATION_SCHEMA':
            return self
        ds = self.index.get(name.upper())
        return ds

    def hasTable(self, tableName):
        tn = tableName.upper()
        if tn in self.information_schema or tn in self.index:
            return True
        return False

    def getTableColumns(self, tableName):
        tn = tableName.upper()
        if tn in self.information_schema:
            return self.information_schema[tn]
        raise Exception()

    def select(self, columns=None, table=None, where=None, order_by=None, group_by=None, came_from=None):
        tn = table.upper()
        if tn == 'SCHEMATA':
            # there is two query we can process, both hardcoded:
            # SELECT schema_name as name FROM INFORMATION_SCHEMA.SCHEMATA;
            # SELECT default_character_set_name as CharacterSetName, default_collation_name as CollationName FROM INFORMATION_SCHEMA.SCHEMATA WHERE schema_name = 'information_schema';
            if len(columns) == 1 and columns[0] == 'schema_name':
                data = [{'schema_name': 'information_schema'}]
                for key in self.index:
                    data.append({
                        'schema_name': key
                    })
                return data
            elif len(columns) == 3 and where is not None and 'schema_name' in where:
                return [{
                    'schema_name': where['schema_name']['$eq'],
                    'default_character_set_name': 'utf8',
                    'default_collation_name': 'utf8_general_ci'
                }]
        if tn == 'TABLES':
            # query examples:
            # SELECT table_name FROM INFORMATION_SCHEMA.TABLES WHERE table_schema = 'information_schema' AND table_type in ('BASE TABLE', 'SYSTEM VIEW');
            # SELECT table_name as name FROM INFORMATION_SCHEMA.TABLES WHERE table_schema = 'information_schema' AND table_type = 'VIEW';
            tables = [
                # at least this tables should be returned for GUI clients
                {'table_name': 'SCHEMATA', 'table_schema': 'information_schema', 'table_type': 'SYSTEM VIEW'},
                {'table_name': 'TABLES', 'table_schema': 'information_schema', 'table_type': 'SYSTEM VIEW'},
                {'table_name': 'EVENTS', 'table_schema': 'information_schema', 'table_type': 'SYSTEM VIEW'},
                {'table_name': 'ROUTINES', 'table_schema': 'information_schema', 'table_type': 'SYSTEM VIEW'},
                {'table_name': 'TRIGGERS', 'table_schema': 'information_schema', 'table_type': 'SYSTEM VIEW'},
            ]
            for dsName, ds in self.index.items():
                t = ds.getTables()
                tables += [{'table_name': x, 'table_schema': dsName, 'table_type': 'BASE TABLE'} for x in t]

            filtered_tables = tables
            if isinstance(where, dict) and 'table_schema' in where:
                schema = where['table_schema']['$eq']
                filtered_tables = [x for x in filtered_tables if x['table_schema'].upper() == schema.upper()]

            if isinstance(where, dict) and 'table_type' in where:
                types = []
                if '$eq' in where['table_type']:
                    types = [where['table_type']['$eq'].upper()]
                if '$in' in where['table_type']:
                    types += [x.upper() for x in where['table_type']['$in']]
                filtered_tables = [x for x in filtered_tables if x['table_type'] in types]

            return filtered_tables
        if tn == 'COLUMNS':
            # SELECT * FROM information_schema.COLUMNS WHERE TABLE_SCHEMA='CSV_DS' AND TABLE_NAME='part' ORDER BY ORDINAL_POSITION
            return []
        if tn == 'EVENTS':
            # SELECT event_name as name FROM INFORMATION_SCHEMA.EVENTS WHERE event_schema = 'information_schema';
            return []
        if tn == 'ROUTINES':
            # SELECT specific_name as name FROM INFORMATION_SCHEMA.ROUTINES WHERE routine_schema = 'information_schema' AND routine_type = 'FUNCTION';
            return []
        if tn == 'TRIGGERS':
            # SELECT trigger_name as name FROM INFORMATION_SCHEMA.TRIGGERS WHERE trigger_schema = 'information_schema';
            return []

        return []



from flask import request
from flask_restx import Resource
from flask import current_app as ca

from mindsdb.api.http.namespaces.configs.util import ns_conf
from mindsdb.utilities.telemetry import (
    enable_telemetry,
    disable_telemetry,
    is_telemetry_file_exists,
    inject_telemetry_to_static
)


@ns_conf.route('/ping')
class Ping(Resource):
    @ns_conf.doc('get_ping')
    def get(self):
        '''Checks server avaliable'''
        return {'status': 'ok'}

@ns_conf.route('/report_uuid')
class ReportUUID(Resource):
    @ns_conf.doc('get_report_uuid')
    def get(self):
        metamodel_name = '___monitroing_metamodel___'
        predictor = ca.mindsdb_native.create(metamodel_name)
        return {
            'report_uuid': predictor.report_uuid
        }

@ns_conf.route('/telemetry')
class Telemetry(Resource):
    @ns_conf.doc('get_telemetry_status')
    def get(self):
        storage_dir = ca.config_obj['storage_dir']
        status = "enabled" if is_telemetry_file_exists(storage_dir) else "disabled"
        return {"status": status}

    @ns_conf.doc('set_telemetry')
    def post(self):
        data = request.json
        action = data['action']
        storage_dir = ca.config_obj['storage_dir']
        if str(action).lower() in ["true", "enable", "on"]:
            enable_telemetry(storage_dir)
        else:
            disable_telemetry(storage_dir)
        inject_telemetry_to_static(ca.config_obj.paths['static'])
        return '', 200



import os
import logging
import multiprocessing
import threading
from pathlib import Path

from werkzeug.exceptions import HTTPException
from waitress import serve
from flask import send_from_directory

from mindsdb.api.http.namespaces.predictor import ns_conf as predictor_ns
from mindsdb.api.http.namespaces.datasource import ns_conf as datasource_ns
from mindsdb.api.http.namespaces.util import ns_conf as utils_ns
from mindsdb.api.http.namespaces.config import ns_conf as conf_ns
from mindsdb.api.http.initialize import initialize_flask, initialize_interfaces, initialize_static
from mindsdb.utilities.config import Config
from mindsdb.utilities.log import initialize_log


def start(config, verbose=False):
    config = Config(config)
    if verbose:
        config['log']['level']['console'] = 'DEBUG'

    initialize_log(config, 'http', wrap_print=True)

    # start static initialization in a separate thread
    init_static_thread = threading.Thread(target=initialize_static, args=(config,))
    init_static_thread.start()

    app, api = initialize_flask(config, init_static_thread)
    initialize_interfaces(config, app)

    static_root = Path(config.paths['static'])

    @app.route('/', defaults={'path': ''}, methods=['GET'])
    @app.route('/<path:path>', methods=['GET'])
    def root_index(path):
        if path.startswith('api/'):
            return {'message': 'wrong query'}, 400
        if static_root.joinpath(path).is_file():
            return send_from_directory(config.paths['static'], path)
        else:
            return send_from_directory(config.paths['static'], 'index.html')

    api.add_namespace(predictor_ns)
    api.add_namespace(datasource_ns)
    api.add_namespace(utils_ns)
    api.add_namespace(conf_ns)

    @api.errorhandler(Exception)
    def handle_exception(e):
        # pass through HTTP errors
        if isinstance(e, HTTPException):
            return {'message': str(e)}, e.code, e.get_response().headers
        name = getattr(type(e), '__name__') or 'Unknown error'
        return {'message': f'{name}: {str(e)}'}, 500

    port = config['api']['http']['port']
    host = config['api']['http']['host']

    server = os.environ.get('MINDSDB_DEFAULT_SERVER', 'waitress')

    # waiting static initialization
    init_static_thread.join()
    if server.lower() == 'waitress':
        serve(app, port=port, host=host)
    elif server.lower() == 'flask':
        # that will 'disable access' log in console
        log = logging.getLogger('werkzeug')
        log.setLevel(logging.WARNING)

        app.run(debug=False, port=port, host=host)
    elif server.lower() == 'gunicorn':
        try:
            from mindsdb.api.http.gunicorn_wrapper import StandaloneApplication
        except ImportError:
            print("Gunicorn server is not available by default. If you wish to use it, please install 'gunicorn'")
            return

        options = {
            'bind': f'{host}:{port}',
            'workers': min(max(multiprocessing.cpu_count(), 2), 3)
        }
        StandaloneApplication(app, options).run()



import mysql.connector

from mindsdb_native.libs.constants.mindsdb import DATA_SUBTYPES
from mindsdb.integrations.base import Integration


class MySQL(Integration):
    def _to_mysql_table(self, stats, predicted_cols):
        subtype_map = {
            DATA_SUBTYPES.INT: 'int',
            DATA_SUBTYPES.FLOAT: 'double',
            DATA_SUBTYPES.BINARY: 'bool',
            DATA_SUBTYPES.DATE: 'Date',
            DATA_SUBTYPES.TIMESTAMP: 'Datetime',
            DATA_SUBTYPES.SINGLE: 'VARCHAR(500)',
            DATA_SUBTYPES.MULTIPLE: 'VARCHAR(500)',
            DATA_SUBTYPES.TAGS: 'VARCHAR(500)',
            DATA_SUBTYPES.IMAGE: 'VARCHAR(500)',
            DATA_SUBTYPES.VIDEO: 'VARCHAR(500)',
            DATA_SUBTYPES.AUDIO: 'VARCHAR(500)',
            DATA_SUBTYPES.SHORT: 'VARCHAR(500)',
            DATA_SUBTYPES.RICH: 'VARCHAR(500)',
            DATA_SUBTYPES.ARRAY: 'VARCHAR(500)'
        }

        column_declaration = []
        for name in stats['columns']:
            try:
                col_subtype = stats[name]['typing']['data_subtype']
                new_type = subtype_map[col_subtype]
                column_declaration.append(f' `{name}` {new_type} ')
                if name in predicted_cols:
                    column_declaration.append(f' `{name}_original` {new_type} ')
            except Exception:
                print(f'Error: cant convert type {col_subtype} of column {name} to mysql type')

        return column_declaration

    def _escape_table_name(self, name):
        return '`' + name.replace('`', '``') + '`'

    def _query(self, query):
        con = mysql.connector.connect(
            host=self.config['integrations'][self.name]['host'],
            port=self.config['integrations'][self.name]['port'],
            user=self.config['integrations'][self.name]['user'],
            password=self.config['integrations'][self.name]['password']
        )

        cur = con.cursor(dictionary=True, buffered=True)
        cur.execute(query)
        res = True
        try:
            res = cur.fetchall()
        except Exception:
            pass
        con.commit()
        con.close()

        return res

    def _get_connect_string(self, table):
        user = f"{self.config['api']['mysql']['user']}_{self.name}"
        password = self.config['api']['mysql']['password']
        host = self.config['api']['mysql']['host']
        port = self.config['api']['mysql']['port']

        if password is None or password == '':
            connect = f'mysql://{user}@{host}:{port}/mindsdb/{table}'
        else:
            connect = f'mysql://{user}:{password}@{host}:{port}/mindsdb/{table}'

        return connect

    def setup(self):
        self._query(f'DROP DATABASE IF EXISTS {self.mindsdb_database}')
        self._query(f'CREATE DATABASE IF NOT EXISTS {self.mindsdb_database}')

        connect = self._get_connect_string('predictors')

        q = f"""
            CREATE TABLE IF NOT EXISTS {self.mindsdb_database}.predictors (
                name VARCHAR(500),
                status VARCHAR(500),
                accuracy VARCHAR(500),
                predict VARCHAR(500),
                select_data_query VARCHAR(500),
                external_datasource VARCHAR(500),
                training_options VARCHAR(500),
                key name_key (name)
            ) ENGINE=FEDERATED CONNECTION='{connect}';
        """
        self._query(q)

        connect = self._get_connect_string('commands')

        q = f"""
            CREATE TABLE IF NOT EXISTS {self.mindsdb_database}.commands (
                command VARCHAR(500),
                key command_key (command)
            ) ENGINE=FEDERATED CONNECTION='{connect}';
        """
        self._query(q)

    def register_predictors(self, model_data_arr):
        for model_meta in model_data_arr:
            name = model_meta['name']
            stats = model_meta['data_analysis_v2']
            columns_sql = ','.join(self._to_mysql_table(stats, model_meta['predict']))
            columns_sql += ',`when_data` varchar(500)'
            columns_sql += ',`select_data_query` varchar(500)'
            columns_sql += ',`external_datasource` varchar(500)'
            for col in model_meta['predict']:
                columns_sql += f',`{col}_confidence` double'
                if model_meta['data_analysis_v2'][col]['typing']['data_type'] == 'Numeric':
                    columns_sql += f',`{col}_min` double'
                    columns_sql += f',`{col}_max` double'
                columns_sql += f',`{col}_explain` varchar(500)'

            connect = self._get_connect_string(name)

            q = f"""
                CREATE TABLE {self.mindsdb_database}.{self._escape_table_name(name)} (
                    {columns_sql},
                    index when_data_index (when_data),
                    index select_data_query_index (select_data_query),
                    index external_datasource_index (external_datasource)
                ) ENGINE=FEDERATED CONNECTION='{connect}';
            """
            self._query(q)

    def unregister_predictor(self, name):
        q = f"""
            drop table if exists {self.mindsdb_database}.{self._escape_table_name(name)};
        """
        self._query(q)

    def check_connection(self):
        try:
            con = mysql.connector.connect(
                host=self.config['integrations'][self.name]['host'],
                port=self.config['integrations'][self.name]['port'],
                user=self.config['integrations'][self.name]['user'],
                password=self.config['integrations'][self.name]['password']
            )
            connected = con.is_connected()
            con.close()
        except Exception:
            connected = False
        return connected



from collections import OrderedDict
from flask_restx import fields
from mindsdb.api.http.namespaces.configs.datasources import ns_conf

get_datasource_rows_params = OrderedDict([
    ('page[size]', {
        'description': 'page size',
        'type': 'integer',
        'in': 'path',
        'required': False,
        'default': 10
    }),
    ('page[offset]', {
        'description': 'start record',
        'type': 'integer',
        'in': 'path',
        'required': False,
        'default': 0
    }),
    ('filter_{type}[{field}]', {
        'description': 'Filter for field with specified type. Type can be one of [like,in,nin,gt,lt,gte,lte,eq,neq]',
        'type': 'string',
        'in': 'path',
        'required': False,
        'default': None
    })
])

datasource_rows_metadata = ns_conf.model('RowsResponse', {
    'data': fields.List(fields.Raw, required=False, description='Filtered content of the datasource'),
    'rowcount': fields.Integer(required=False, description='Count of rows in dataset'),
    'columns_names': fields.List(fields.String, required=False, description='The name of the columns')
})

EXAMPLES = [{
    'data': [
        {
            'name': 'Mercury',
            'mass': 0.055,
            'radius': 0.3829
        }, {
            'name': 'Venus',
            'mass': 0.815,
            'radius': 0.9499
        }, {
            'name': 'Earth',
            'mass': 1.0,
            'radius': 1.0
        }
    ]
}]



# Cochran’s sample size calculator

def sampleSize(
    population_size,
    margin_error=.05,
    confidence_level=.99,
    sigma=1/2
):
    """
    Calculate the minimal sample size to use to achieve a certain
    margin of error and confidence level for a sample estimate
    of the population mean.
    Inputs
    -------
    population_size: integer
        Total size of the population that the sample is to be drawn from.
    margin_error: number
        Maximum expected difference between the true population parameter,
        such as the mean, and the sample estimate.
    confidence_level: number in the interval (0, 1)
        If we were to draw a large number of equal-size samples
        from the population, the true population parameter
        should lie within this percentage
        of the intervals (sample_parameter - e, sample_parameter + e)
        where e is the margin_error.
    sigma: number
        The standard deviation of the population.  For the case
        of estimating a parameter in the interval [0, 1], sigma=1/2
        should be sufficient.
    """
    alpha = 1 - (confidence_level)
    # dictionary of confidence levels and corresponding z-scores
    # computed via norm.ppf(1 - (alpha/2)), where norm is
    # a normal distribution object in scipy.stats.
    # Here, ppf is the percentile point function.
    zdict = {
        .90: 1.645,
        .91: 1.695,
        .99: 2.576,
        .97: 2.17,
        .94: 1.881,
        .93: 1.812,
        .95: 1.96,
        .98: 2.326,
        .96: 2.054,
        .92: 1.751
    }
    if confidence_level in zdict:
        z = zdict[confidence_level]
    else:
        from scipy.stats import norm
        z = norm.ppf(1 - (alpha/2))
    N = population_size
    M = margin_error
    numerator = z**2 * sigma**2 * (N / (N-1))
    denom = M**2 + ((z**2 * sigma**2)/(N-1))
    return numerator/denom



import os
import shutil
import importlib
import json
import sys

import mindsdb_native
import pandas as pd

from mindsdb.interfaces.database.database import DatabaseWrapper
from mindsdb.interfaces.native.native import NativeInterface

class CustomModels():
    def __init__(self, config):
        self.config = config
        self.dbw = DatabaseWrapper(self.config)
        self.storage_dir = os.path.join(config['storage_dir'], 'misc')
        os.makedirs(self.storage_dir, exist_ok=True)
        self.model_cache = {}
        self.mindsdb_native = NativeInterface(self.config)
        self.dbw = DatabaseWrapper(self.config)

    def _dir(self, name):
        return str(os.path.join(self.storage_dir, 'custom_model_' + name))

    def _internal_load(self, name):

        # Caching (2 lines bellow), currently disabled due to multiprocessing cache invalidation issues
        #if name in self.model_cache:
        #    return self.model_cache[name]

        # "Proper" model loading (3 lines bellow), currently disabled due to pickling issues
        #spec = importlib.util.spec_from_file_location(name, self._dir(name) + '/model.py')
        #module = importlib.util.module_from_spec(spec)
        #spec.loader.exec_module(module)

        sys.path.insert(0, self._dir(name))
        module = __import__(name)

        try:
            model = module.Model.load(os.path.join(self._dir(name), 'model.pickle'))
        except Exception as e:
            model = module.Model()
            model.initialize_column_types()
            if hasattr(model, 'setup'):
                model.setup()

        self.model_cache[name] = model

        return model

    def learn(self, name, from_data, to_predict, kwargs={}):
        model_data = self.get_model_data(name)
        model_data['status'] = 'training'
        self.save_model_data(name, model_data)

        to_predict = to_predict if isinstance(to_predict,list) else [to_predict]
        data_source = getattr(mindsdb_native, from_data['class'])(*from_data['args'], **from_data['kwargs'])
        data_frame = data_source.df
        model = self._internal_load(name)
        model.to_predict = to_predict

        model_data = self.get_model_data(name)
        model_data['predict'] = model.to_predict
        self.save_model_data(name, model_data)

        data_analysis = self.mindsdb_native.analyse_dataset(data_source)['data_analysis_v2']

        model_data = self.get_model_data(name)
        model_data['data_analysis_v2'] = data_analysis
        self.save_model_data(name, model_data)

        model.fit(data_frame, to_predict, data_analysis, kwargs)

        model.save(os.path.join(self._dir(name), 'model.pickle'))
        self.model_cache[name] = model

        model_data = self.get_model_data(name)
        model_data['status'] = 'completed'
        self.save_model_data(name, model_data)

        self.dbw.unregister_predictor(name)
        self.dbw.register_predictors([self.get_model_data(name)])

    def predict(self, name, when_data=None, from_data=None, kwargs=None):
        if kwargs is None:
            kwargs = {}
        if from_data is not None:
            if isinstance(from_data, dict):
                data_source = getattr(mindsdb_native, from_data['class'])(*from_data['args'],
                                                                          **from_data['kwargs'])
            # assume that particular instance of any DataSource class is provided
            else:
                data_source = from_data
            data_frame = data_source.df
        elif when_data is not None:
            if isinstance(when_data, dict):
                for k in when_data:
                    when_data[k] = [when_data[k]]
                data_frame = pd.DataFrame(when_data)
            else:
                data_frame = pd.DataFrame(when_data)

        model = self._internal_load(name)
        predictions = model.predict(data_frame, kwargs)

        pred_arr = []
        for i in range(len(predictions)):
            pred_arr.append({})
            pred_arr[-1] = {}
            for col in predictions.columns:
                pred_arr[-1][col] = {}
                pred_arr[-1][col]['predicted_value'] = predictions[col].iloc[i]

        return pred_arr

    def get_model_data(self, name):
        with open(os.path.join(self._dir(name), 'metadata.json'), 'r') as fp:
            return json.load(fp)

    def save_model_data(self, name, data):
        with open(os.path.join(self._dir(name), 'metadata.json'), 'w') as fp:
            json.dump(data, fp)

    def get_models(self):
        models = []
        for model_dir in os.listdir(self.storage_dir):
            if 'custom_model_' in model_dir:
                name = model_dir.replace('custom_model_','')
                try:
                    models.append(self.get_model_data(name))
                except:
                    print(f'Model {name} not found !')

        return models

    def delete_model(self, name):
        shutil.rmtree(self._dir(name))
        self.dbw.unregister_predictor(name)

    def rename_model(self, name, new_name):
        self.dbw.unregister_predictor(name)
        shutil.move(self._dir(name), self._dir(new_name))
        shutil.move(os.path.join(self._dir(new_name) + f'{name}.py'), os.path.join(self._dir(new_name) ,f'{new_name}.py'))
        self.dbw.register_predictors([self.get_model_data(new_name)])

    def export_model(self, name):
        shutil.make_archive(base_name=name, format='zip', root_dir=self._dir(name))
        return str(self._dir(name)) + '.zip'

    def load_model(self, fpath, name, trained_status):
        shutil.unpack_archive(fpath, self._dir(name), 'zip')
        shutil.move( os.path.join(self._dir(name), 'model.py') ,  os.path.join(self._dir(name), f'{name}.py') )
        model = self._internal_load(name)
        model.to_predict = model.to_predict if isinstance(model.to_predict,list) else [model.to_predict]
        self.save_model_data(name,{
            'name': name
            ,'data_analysis_v2': model.column_type_map
            ,'predict': model.to_predict
            ,'status': trained_status
            ,'is_custom': True
        })

        with open(os.path.join(self._dir(name), '__init__.py') , 'w') as fp:
            fp.write('')

        if trained_status == 'trained':
            self.dbw.register_predictors([self.get_model_data(name)])



from mindsdb.api.http.namespaces.configs.predictors import ns_conf

from flask_restx import fields
import datetime

predictor_status = ns_conf.model('PredictorStatus', {
    # Primary key
    'name': fields.String(required=False, description='The predictor name, NOTE: That primary key is made of name:version'),
    'version': fields.String(required=False, description='The predictor version to publish under, this is so that we can train multiple predictors for the same problem but expose them via the same name'),
    # other attributes
    'is_active': fields.Boolean(required=False, description='Only one predictor by public_name can be active'),
    'data_source': fields.String(required=False, description='The data source it\'s learning from'),
    'predict': fields.List(fields.String, required=False, description='The list of columns/fields to be predicted'),
    'accuracy': fields.Float(description='The current accuracy of the model'),
    'status': fields.String(required=False, description='The current model status', enum=['training', 'complete', 'error']),
    'current_phase': fields.String(required=False, description='Current training phase'),
    'train_end_at': fields.DateTime(required=False, description='The time the predictor finished training'),
    'updated_at': fields.DateTime(required=False, description='The time the predictor was last updated at'),
    'created_at': fields.DateTime(required=False, description='The time the predictor was created at'),
    'is_custom': fields.Boolean(default=False),
})




##EXAMPLES



EXAMPLES = [
    {
        'name': 'Price',
        'version': 1,
        'is_active': True,
        'data_source': 'real_estate.csv',
        'predict': ['price'],
        'accuracy': '.97',
        'status': 'training',
        'train_end_at': datetime.datetime.now(),
        'updated_at': datetime.datetime.now(),
        'created_at': datetime.datetime.now()
    },
    {
        'name': 'Price',
        'version': 2,
        'is_active': False,
        'data_source': 'real_estate_complex.csv',
        'predict': ['price'],
        'accuracy': '.64',
        'status': 'training',
        'train_end_at': datetime.datetime.now(),
        'updated_at': datetime.datetime.now(),
        'created_at': datetime.datetime.now()
    },
    {
        'name': 'Number of rooms',
        'version': 1,
        'is_active': True,
        'data_source': 'real_estate.csv',
        'predict': ['number_of_rooms'],
        'accuracy': '.97',
        'status': 'complete',
        'train_end_at': datetime.datetime.now(),
        'updated_at': datetime.datetime.now(),
        'created_at': datetime.datetime.now()
    }
]



import requests

from mindsdb_native.libs.constants.mindsdb import DATA_SUBTYPES
from mindsdb.integrations.base import Integration


class Clickhouse(Integration):
    def _to_clickhouse_table(self, stats, predicted_cols):
        subtype_map = {
            DATA_SUBTYPES.INT: 'Nullable(Int64)',
            DATA_SUBTYPES.FLOAT: 'Nullable(Float64)',
            DATA_SUBTYPES.BINARY: 'Nullable(UInt8)',
            DATA_SUBTYPES.DATE: 'Nullable(Date)',
            DATA_SUBTYPES.TIMESTAMP: 'Nullable(Datetime)',
            DATA_SUBTYPES.SINGLE: 'Nullable(String)',
            DATA_SUBTYPES.MULTIPLE: 'Nullable(String)',
            DATA_SUBTYPES.TAGS: 'Nullable(String)',
            DATA_SUBTYPES.IMAGE: 'Nullable(String)',
            DATA_SUBTYPES.VIDEO: 'Nullable(String)',
            DATA_SUBTYPES.AUDIO: 'Nullable(String)',
            DATA_SUBTYPES.SHORT: 'Nullable(String)',
            DATA_SUBTYPES.RICH: 'Nullable(String)',
            DATA_SUBTYPES.ARRAY: 'Nullable(String)'
        }

        column_declaration = []
        for name in stats['columns']:
            try:
                col_subtype = stats[name]['typing']['data_subtype']
                new_type = subtype_map[col_subtype]
                column_declaration.append(f' `{name}` {new_type} ')
                if name in predicted_cols:
                    column_declaration.append(f' `{name}_original` {new_type} ')
            except Exception as e:
                print(e)
                print(f'Error: cant convert type {col_subtype} of column {name} to clickhouse type')

        return column_declaration

    def _query(self, query):
        params = {'user': 'default'}
        try:
            params['user'] = self.config['integrations'][self.name]['user']
        except Exception:
            pass

        try:
            params['password'] = self.config['integrations'][self.name]['password']
        except Exception:
            pass

        host = self.config['integrations'][self.name]['host']
        port = self.config['integrations'][self.name]['port']

        response = requests.post(f'http://{host}:{port}', data=query, params=params)

        if response.status_code != 200:
            raise Exception(f'Error: {response.content}\nQuery:{query}')

        return response

    def _get_mysql_user(self):
        return f"{self.config['api']['mysql']['user']}_{self.name}"

    def _escape_table_name(self, name):
        return '`' + name.replace('`', '\\`') + '`'

    def setup(self):
        self._query(f'DROP DATABASE IF EXISTS {self.mindsdb_database}')
        self._query(f'CREATE DATABASE IF NOT EXISTS {self.mindsdb_database}')

        msqyl_conn = self.config['api']['mysql']['host'] + ':' + str(self.config['api']['mysql']['port'])
        msqyl_pass = self.config['api']['mysql']['password']
        msqyl_user = self._get_mysql_user()

        q = f"""
            CREATE TABLE IF NOT EXISTS {self.mindsdb_database}.predictors (
                name String,
                status String,
                accuracy String,
                predict String,
                select_data_query String,
                external_datasource String,
                training_options String
                ) ENGINE=MySQL('{msqyl_conn}', 'mindsdb', 'predictors', '{msqyl_user}', '{msqyl_pass}')
        """
        self._query(q)
        q = f"""
            CREATE TABLE IF NOT EXISTS {self.mindsdb_database}.commands (
                command String
            ) ENGINE=MySQL('{msqyl_conn}', 'mindsdb', 'commands', '{msqyl_user}', '{msqyl_pass}')
        """
        self._query(q)

    def register_predictors(self, model_data_arr):
        for model_meta in model_data_arr:
            name = self._escape_table_name(model_meta['name'])
            stats = model_meta['data_analysis_v2']

            columns_sql = ','.join(self._to_clickhouse_table(stats, model_meta['predict']))
            columns_sql += ',`when_data` Nullable(String)'
            columns_sql += ',`select_data_query` Nullable(String)'
            columns_sql += ',`external_datasource` Nullable(String)'
            for col in model_meta['predict']:
                columns_sql += f',`{col}_confidence` Nullable(Float64)'

                if model_meta['data_analysis_v2'][col]['typing']['data_type'] == 'Numeric':
                    columns_sql += f',`{col}_min` Nullable(Float64)'
                    columns_sql += f',`{col}_max` Nullable(Float64)'
                columns_sql += f',`{col}_explain` Nullable(String)'

            msqyl_conn = self.config['api']['mysql']['host'] + ':' + str(self.config['api']['mysql']['port'])
            msqyl_pass = self.config['api']['mysql']['password']
            msqyl_user = self._get_mysql_user()

            q = f"""
                CREATE TABLE {self.mindsdb_database}.{name}
                ({columns_sql}
                ) ENGINE=MySQL('{msqyl_conn}', 'mindsdb', {name}, '{msqyl_user}', '{msqyl_pass}')
            """
            self._query(q)

    def unregister_predictor(self, name):
        q = f"""
            drop table if exists {self.mindsdb_database}.{self._escape_table_name(name)};
        """
        self._query(q)

    def check_connection(self):
        try:
            res = self._query('select 1;')
            connected = res.status_code == 200
        except Exception:
            connected = False
        return connected



# Mindsdb native interface
from pathlib import Path
import json
import datetime
from dateutil.parser import parse as parse_datetime
import psutil

import mindsdb_native
from mindsdb_native import F
from mindsdb.utilities.fs import create_directory
from mindsdb_native.libs.constants.mindsdb import DATA_SUBTYPES
from mindsdb.interfaces.native.learn_process import LearnProcess
from mindsdb.interfaces.database.database import DatabaseWrapper


class NativeInterface():
    def __init__(self, config):
        self.config = config
        self.dbw = DatabaseWrapper(self.config)
        self.predictor_cache = {}

    def _invalidate_cached_predictors(self):
        # @TODO: Cache will become stale if the respective NativeInterface is not invoked yet a bunch of predictors remained cached, no matter where we invoke it. In practice shouldn't be a big issue though
        for predictor_name in list(self.predictor_cache.keys()):
            if (datetime.datetime.now() - self.predictor_cache[predictor_name]['created']).total_seconds() > 1200:
                del self.predictor_cache[predictor_name]

    def _setup_for_creation(self, name):
        if name in self.predictor_cache:
            del self.predictor_cache[name]
        # Here for no particular reason, because we want to run this sometimes but not too often
        self._invalidate_cached_predictors()

        predictor_dir = Path(self.config.paths['predictors']).joinpath(name)
        create_directory(predictor_dir)
        versions_file_path = predictor_dir.joinpath('versions.json')
        with open(str(versions_file_path), 'wt') as f:
            json.dump(self.config.versions, f, indent=4, sort_keys=True)

    def create(self, name):
        self._setup_for_creation(name)
        predictor = mindsdb_native.Predictor(name=name, run_env={'trigger': 'mindsdb'})
        return predictor

    def learn(self, name, from_data, to_predict, kwargs={}):
        join_learn_process = kwargs.get('join_learn_process', False)
        if 'join_learn_process' in kwargs:
            del kwargs['join_learn_process']

        self._setup_for_creation(name)

        p = LearnProcess(name, from_data, to_predict, kwargs, self.config.get_all())
        p.start()
        if join_learn_process is True:
            p.join()
            if p.exitcode != 0:
                raise Exception('Learning process failed !')

    def predict(self, name, when_data=None, kwargs={}):
        if name not in self.predictor_cache:
            # Clear the cache entirely if we have less than .12 GB left
            if psutil.virtual_memory().available < 1.2 * pow(10,9):
                self.predictor_cache = {}

            if F.get_model_data(name)['status'] == 'complete':
                self.predictor_cache[name] = {
                    'predictor': mindsdb_native.Predictor(name=name, run_env={'trigger': 'mindsdb'}),
                    'created': datetime.datetime.now()
                }

        predictions = self.predictor_cache[name]['predictor'].predict(
            when_data=when_data,
            **kwargs
        )

        return predictions

    def analyse_dataset(self, ds):
        return F.analyse_dataset(ds)

    def get_model_data(self, name, db_fix=True):
        model = F.get_model_data(name)

        # Make some corrections for databases not to break when dealing with empty columns
        if db_fix:
            data_analysis = model['data_analysis_v2']
            for column in data_analysis['columns']:
                analysis = data_analysis.get(column)
                if isinstance(analysis, dict) and (len(analysis) == 0 or analysis.get('empty', {}).get('is_empty', False)):
                    data_analysis[column]['typing'] = {
                        'data_subtype': DATA_SUBTYPES.INT
                    }

        return model

    def get_models(self):
        models = []
        predictors = [
            x for x in Path(self.config.paths['predictors']).iterdir() if
                x.is_dir()
                and x.joinpath('light_model_metadata.pickle').is_file()
                and x.joinpath('heavy_model_metadata.pickle').is_file()
        ]
        for p in predictors:
            model_name = p.name
            try:
                model_data = self.get_model_data(model_name, db_fix=False)
                if model_data['status'] == 'training' and parse_datetime(model_data['created_at']) < parse_datetime(self.config['mindsdb_last_started_at']):
                    continue

                reduced_model_data = {}

                for k in ['name', 'version', 'is_active', 'predict', 'status', 'current_phase', 'accuracy', 'data_source']:
                    reduced_model_data[k] = model_data.get(k, None)

                for k in ['train_end_at', 'updated_at', 'created_at']:
                    reduced_model_data[k] = model_data.get(k, None)
                    if reduced_model_data[k] is not None:
                        try:
                            reduced_model_data[k] = parse_datetime(str(reduced_model_data[k]).split('.')[0])
                        except Exception as e:
                            # @TODO Does this ever happen
                            print(f'Date parsing exception while parsing: {k} in get_models: ', e)
                            reduced_model_data[k] = parse_datetime(str(reduced_model_data[k]))

                models.append(reduced_model_data)
            except Exception as e:
                print(f"Can't list data for model: '{model_name}' when calling `get_models(), error: {e}`")

        return models

    def delete_model(self, name):
        F.delete_model(name)
        self.dbw.unregister_predictor(name)

    def rename_model(self, name, new_name):
        self.dbw.unregister_predictor(self.get_model_data(name))
        F.rename_model(name, new_name)
        self.dbw.register_predictors(self.get_model_data(new_name))

    def load_model(self, fpath):
        name = F.import_model(model_archive_path=fpath)
        self.dbw.register_predictors(self.get_model_data(name), setup=False)

    def export_model(self, name):
        F.export_predictor(model_name=name)



import copy
import traceback

from flask import request
from flask_restx import Resource, abort
from flask import current_app as ca

from mindsdb.api.http.namespaces.configs.config import ns_conf
from mindsdb.utilities.functions import get_all_models_meta_data


def get_integration(name):
    integrations = ca.config_obj.get('integrations', {})
    return integrations.get(name)


@ns_conf.route('/integrations')
@ns_conf.param('name', 'List all database integration')
class ListIntegration(Resource):
    @ns_conf.doc('get_integrations')
    def get(self):
        return {'integrations': [k for k in ca.config_obj.get('integrations', {})]}


@ns_conf.route('/all_integrations')
@ns_conf.param('name', 'List all database integration')
class AllIntegration(Resource):
    @ns_conf.doc('get_all_integrations')
    def get(self):
        integrations = copy.deepcopy(
            ca.config_obj.get('integrations', {})
        )
        for integration in integrations.values():
            if 'password' in integration:
                integration['password'] = None
        return integrations


@ns_conf.route('/integrations/<name>')
@ns_conf.param('name', 'Database integration')
class Integration(Resource):
    @ns_conf.doc('get_integration')
    def get(self, name):
        integration = get_integration(name)
        if integration is None:
            abort(404, f'Can\'t find database integration: {name}')
        integration = copy.deepcopy(integration)
        if 'password' in integration:
            integration['password'] = None
        return integration

    @ns_conf.doc('put_integration')
    def put(self, name):
        params = request.json.get('params')
        if not isinstance(params, dict):
            abort(400, "type of 'params' must be dict")

        is_test = params.get('test', False)
        if is_test:
            del params['test']

        integration = get_integration(name)
        if integration is not None:
            abort(400, f"Integration with name '{name}' already exists")
        try:
            if 'enabled' in params:
                params['publish'] = params['enabled']
                del params['enabled']
            ca.config_obj.add_db_integration(name, params)

            mdb = ca.mindsdb_native
            cst = ca.custom_models
            model_data_arr = get_all_models_meta_data(mdb, cst)
            ca.dbw.setup_integration(name)
            if is_test is False:
                ca.dbw.register_predictors(model_data_arr)
        except Exception as e:
            print(traceback.format_exc())
            abort(500, f'Error during config update: {str(e)}')

        if is_test:
            cons = ca.dbw.check_connections()
            ca.config_obj.remove_db_integration(name)
            return {'success': cons[name]}, 200

        return '', 200

    @ns_conf.doc('delete_integration')
    def delete(self, name):
        integration = get_integration(name)
        if integration is None:
            abort(400, f"Nothing to delete. '{name}' not exists.")
        try:
            ca.config_obj.remove_db_integration(name)
        except Exception as e:
            print(traceback.format_exc())
            abort(500, f'Error during integration delete: {str(e)}')
        return '', 200

    @ns_conf.doc('modify_integration')
    def post(self, name):
        params = request.json.get('params')
        if not isinstance(params, dict):
            abort(400, "type of 'params' must be dict")
        integration = get_integration(name)
        if integration is None:
            abort(400, f"Nothin to modify. '{name}' not exists.")
        try:
            if 'enabled' in params:
                params['publish'] = params['enabled']
                del params['enabled']
            ca.config_obj.modify_db_integration(name, params)
            ca.dbw.setup_integration(name)
        except Exception as e:
            print(traceback.format_exc())
            abort(500, f'Error during integration modifycation: {str(e)}')
        return '', 200


@ns_conf.route('/integrations/<name>/check')
@ns_conf.param('name', 'Database integration checks')
class Check(Resource):
    @ns_conf.doc('check')
    def get(self, name):
        if get_integration(name) is None:
            abort(404, f'Can\'t find database integration: {name}')
        connections = ca.dbw.check_connections()
        return connections.get(name, False), 200


@ns_conf.route('/telemetry/<flag>')
@ns_conf.param('flag', 'Turn telemtry on or off')
class ToggleTelemetry(Resource):
    @ns_conf.doc('check')
    def get(self, flag):
        if flag in ["True", "true", "t"]:
            return 'Enabled telemetry', 200
        else:
            return 'Disabled telemetry', 200



from mindsdb.api.http.namespaces.configs.predictors import ns_conf

from flask_restx import fields


data_preparation_metadata = ns_conf.model('DataPreparationMetadata', {
    'accepted_margin_of_error': fields.Float(required=False, description='This is the margin of error that the user accepted when training the predictor, based on this we estimate how much data to actually sample from the provided data set'),
    'total_row_count': fields.Integer(required=False, description='The total number of rows found on the data set'),
    'used_row_count': fields.Integer(required=False, description='The number of rows sampled fro the entire dataset, this is calculated accordingly from the margin of error argument'),
    'test_row_count': fields.Integer(required=False, description='The number of rows used on the test subset'),
    'train_row_count': fields.Integer(required=False, description='The number of rows used on the train subset'),
    'validation_row_count': fields.Integer(required=False, description='The number of rows used on the validation subset')
})


EXAMPLE = {
    'accepted_margin_of_error': 0.2,
    'total_row_count': 18000,
    'used_row_count': 10000,
    'test_row_count': 1000,
    'validation_row_count': 1000,
    'train_row_count': 8000
}



import os
import json
import hashlib
import datetime


default_config = {
    "log": {
        "level": {
            "console": "ERROR",
            "file": "WARNING"
        }

    },
    "debug": False,
    "integrations": {},
    "api": {
        "http": {
            "host": "127.0.0.1",
            "port": "47334"
        },
        "mysql": {
            "host": "127.0.0.1",
            "password": "",
            "port": "47335",
            "user": "mindsdb",
            "database": "mindsdb",
            "ssl": True
        },
        "mongodb": {
            "host": "127.0.0.1",
            "port": "47336",
            "database": "mindsdb"
        }
    }
}


class Config(object):
    current_version = '1.4'
    _config = {}
    paths = {
        'root': '',
        'datasources': '',
        'predictors': '',
        'static': '',
        'tmp': '',
        'log': '',
        'obsolete': {
            'predictors': '',
            'datasources': ''
        }
    }
    versions = {}

    def __init__(self, config_path):
        self._config_path = None
        self._config_hash = None
        self._config = None
        if isinstance(config_path, str):
            self.config_path = os.path.abspath(config_path)
            self._read()
            self._config_hash = self._gen_hash()

            storage_dir = self._config['storage_dir']
            if os.path.isabs(storage_dir) is False:
                storage_dir = os.path.normpath(
                    os.path.join(
                        os.path.dirname(self.config_path),
                        storage_dir
                    )
                )
            self.paths['root'] = storage_dir
            self.paths['datasources'] = os.path.join(storage_dir, 'datasources')
            self.paths['predictors'] = os.path.join(storage_dir, 'predictors')
            self.paths['static'] = os.path.join(storage_dir, 'static')
            self.paths['tmp'] = os.path.join(storage_dir, 'tmp')
            self.paths['log'] = os.path.join(storage_dir, 'log')
            self.paths['obsolete']['predictors'] = os.path.join(storage_dir, 'obsolete', 'predictors')
            self.paths['obsolete']['datasources'] = os.path.join(storage_dir, 'obsolete', 'datasources')

            self._read_versions_file(os.path.join(self.paths['root'], 'versions.json'))
        else:
            raise TypeError('Argument must be string representing a file path <Later on to be switched to file path and/or database connection info>')

    def _read_versions_file(self, path):
        if os.path.isfile(path):
            with open(path, 'rt') as f:
                self.versions = json.loads(f.read())

    def _migrate(self):
        def m1_0(config):
            if 'default_clickhouse' in config['integrations'] and 'type' not in config['integrations']['default_clickhouse']:
                config['integrations']['default_clickhouse']['type'] = 'clickhouse'
            if 'default_mariadb' in config['integrations'] and 'type' not in config['integrations']['default_mariadb']:
                config['integrations']['default_mariadb']['type'] = 'mariadb'
            if 'datasources' in config['api']['mysql']:
                del config['api']['mysql']['datasources']
            config['config_version'] = '1.1'
            return config

        def m1_1(config):
            import tempfile
            import shutil
            from pathlib import Path

            ds_storage_path = Path(config['interface']['datastore']['storage_dir'])
            mdb_storage_path = Path(config['interface']['mindsdb_native']['storage_dir'])

            temp_dir_path = tempfile.mkdtemp()

            if ds_storage_path.is_dir():
                shutil.move(
                    str(ds_storage_path),
                    temp_dir_path
                )

            ds_storage_path.mkdir(mode=0o777, exist_ok=True, parents=True)

            if Path(temp_dir_path).joinpath('datastore').is_dir():
                shutil.move(
                    str(Path(temp_dir_path).joinpath('datastore')),
                    str(ds_storage_path.joinpath('datasources'))
                )
            else:
                ds_storage_path.joinpath('datasources').mkdir(mode=0o777, exist_ok=True)

            if ds_storage_path == mdb_storage_path:
                shutil.move(
                    str(Path(temp_dir_path)),
                    str(ds_storage_path.joinpath('predictors'))
                )
            elif mdb_storage_path.is_dir():
                shutil.move(
                    str(mdb_storage_path),
                    str(ds_storage_path.joinpath('predictors'))
                )
            else:
                mdb_storage_path.joinpath('predictors').mkdir(mode=0o777, exist_ok=True)

            ds_storage_path.joinpath('tmp').mkdir(mode=0o777, exist_ok=True)
            ds_storage_path.joinpath('static').mkdir(mode=0o777, exist_ok=True)

            if Path(temp_dir_path).is_dir():
                shutil.rmtree(temp_dir_path)

            config['storage_dir'] = str(ds_storage_path)
            del config['interface']['datastore']['storage_dir']
            del config['interface']['mindsdb_native']['storage_dir']
            config['config_version'] = '1.2'
            return config

        def m1_2(config):
            ''' remove no longer needed fields
            '''
            try:
                del config['api']['mysql']['log']
            except Exception:
                pass

            try:
                del config['interface']
            except Exception:
                pass

            if 'pip_path' in config and config['pip_path'] is None:
                del config['pip_path']

            if 'python_interpreter' in config and config['python_interpreter'] is None:
                del config['python_interpreter']

            config['config_version'] = '1.3'
            return config

        def m1_3(config):
            ''' rename integration['enabled'] to integration['publish']
            '''
            for integration in config.get('integrations', []).values():
                if 'enabled' in integration:
                    enabled = integration['enabled']
                    del integration['enabled']
                    integration['publish'] = enabled

            config['config_version'] = '1.4'
            return config

        migrations = {
            '1.0': m1_0,
            '1.1': m1_1,
            '1.2': m1_2,
            '1.3': m1_3
        }

        current_version = self._parse_version(self._config['config_version'])
        target_version = self._parse_version(self.current_version)
        while current_version < target_version:
            str_version = '.'.join([str(x) for x in current_version])
            self._config = migrations[str_version](self._config)
            current_version = self._parse_version(self._config['config_version'])

    def _validate(self):
        integrations = self._config.get('integrations', {})
        for key, value in integrations.items():
            if not isinstance(value, dict):
                raise TypeError(f"Config error: integration '{key}' must be a json")
            if 'type' not in integrations[key]:
                raise KeyError(f"Config error: for integration '{key}' key 'type' must be specified")

        storage_dir = self._config.get('storage_dir')
        if storage_dir is None:
            raise KeyError("'storage_dir' mandatory key in config")

    def _parse_version(self, version):
        if isinstance(version, str):
            version = [int(x) for x in version.split('.')]
        elif isinstance(version, int):
            version = [version]
        if len(version) == 1:
            version.append(0)
        return version

    def _format(self):
        ''' changing user input to formalised view
        '''
        for integration in self._config.get('integrations', {}).values():
            password = integration.get('password')
            password = '' if password is None else str(password)
            integration['password'] = str(password)

        password = self._config['api']['mysql'].get('password')
        password = '' if password is None else str(password)
        self._config['api']['mysql']['password'] = str(password)

    def _merge_default_config(self):
        def merge_key_recursive(target_dict, source_dict, key):
            if key not in target_dict:
                target_dict[key] = source_dict[key]
            elif isinstance(target_dict[key], dict) and isinstance(source_dict[key], dict):
                for k in source_dict[key]:
                    merge_key_recursive(target_dict[key], source_dict[key], k)

        for key in default_config:
            merge_key_recursive(self._config, default_config, key)

    def _read(self):
        if isinstance(self.config_path, str) and os.path.isfile(self.config_path):
            with open(self.config_path, 'r') as fp:
                self._config = json.load(fp)
                if self._parse_version(self._config['config_version']) < self._parse_version(self.current_version):
                    self._migrate()
                    self._save()
                self._validate()
                self._merge_default_config()
                self._format()
        else:
            raise TypeError('`self.config_path` must be a string representing a local file path to a json config')

    def _save(self):
        with open(self.config_path, 'w') as fp:
            json.dump(self._config, fp, indent=4, sort_keys=True)

    def _gen_hash(self):
        with open(self.config_path, 'rb') as fp:
            return hashlib.md5(fp.read()).hexdigest()

    def _set_updated(self, key):
        # Only check this for dynamically generated keys, won't be needed once we switch to using a database here
        if key in ['integrations']:
            file_hash = self._gen_hash()
            if file_hash != self._config_hash:
                self._read()
                self._config_hash = self._gen_hash()

    def __getitem__(self, key):
        self._set_updated(key)
        return self._config[key]

    def get(self, key, default=None):
        self._set_updated(key)
        return self._config.get(key, default)

    def get_all(self):
        return self._config

    def set(self, key_chain, value, delete=False):
        self._read()
        c = self._config
        for i, k in enumerate(key_chain):
            if k in c and i + 1 < len(key_chain):
                c = c[k]
            elif k not in c and i + 1 < len(key_chain):
                c[k] = {}
                c = c[k]
            else:
                if delete:
                    del c[k]
                else:
                    c[k] = value
        self._save()

    # Higher level interface
    def add_db_integration(self, name, dict):
        dict['date_last_update'] = str(datetime.datetime.now()).split('.')[0]
        if 'database_name' not in dict:
            dict['database_name'] = name
        if 'publish' not in dict:
            dict['publish'] = True

        self.set(['integrations', name], dict)

    def modify_db_integration(self, name, dict):
        old_dict = self._config['integrations'][name]
        for k in old_dict:
            if k not in dict:
                dict[k] = old_dict[k]

        self.add_db_integration(name, dict)

    def remove_db_integration(self, name):
        self.set(['integrations', name], None, True)



from abc import ABC, abstractmethod


class Integration(ABC):
    def __init__(self, config, name):
        self.config = config
        self.name = name
        self.mindsdb_database = config['api']['mysql']['database']

    @abstractmethod
    def setup(self):
        pass

    @abstractmethod
    def _query(self, query, fetch=False):
        pass

    @abstractmethod
    def register_predictors(self, model_data_arr):
        pass

    @abstractmethod
    def unregister_predictor(self, name):
        pass

    @abstractmethod
    def check_connection(self):
        pass



from distutils.version import LooseVersion
import requests
import os
import shutil
import threading
import webbrowser
from zipfile import ZipFile
from pathlib import Path
import logging
import traceback
#import concurrent.futures

from flask import Flask, url_for
from flask_restx import Api

from mindsdb.__about__ import __version__ as mindsdb_version
from mindsdb.interfaces.datastore.datastore import DataStore
from mindsdb.interfaces.native.native import NativeInterface
from mindsdb.interfaces.custom.custom_models import CustomModels
from mindsdb.utilities.ps import is_pid_listen_port, wait_func_is_true
from mindsdb.interfaces.database.database import DatabaseWrapper
from mindsdb.utilities.telemetry import inject_telemetry_to_static


class Swagger_Api(Api):
    """
    This is a modification of the base Flask Restplus Api class due to the issue described here
    https://github.com/noirbizarre/flask-restplus/issues/223
    """
    @property
    def specs_url(self):
        return url_for(self.endpoint("specs"), _external=False)


def initialize_static(config):
    ''' Update Scout files basing on compatible-config.json content.
        Files will be downloaded and updated if new version of GUI > current.
        Current GUI version stored in static/version.txt.
    '''
    log = logging.getLogger('mindsdb.http')
    static_path = Path(config.paths['static'])
    static_path.mkdir(parents=True, exist_ok=True)

    try:
        res = requests.get('https://mindsdb-web-builds.s3.amazonaws.com/compatible-config.json')
    except (ConnectionError, requests.exceptions.ConnectionError) as e:
        print(f'Is no connection. {e}')
        return False
    except Exception as e:
        print(f'Is something wrong with getting compatible-config.json: {e}')
        return False

    if res.status_code != 200:
        print(f'Cant get compatible-config.json: returned status code = {res.status_code}')
        return False

    try:
        versions = res.json()
    except Exception as e:
        print(f'Cant decode compatible-config.json: {e}')
        return False

    current_mindsdb_lv = LooseVersion(mindsdb_version)

    try:
        gui_versions = {}
        max_mindsdb_lv = None
        max_gui_lv = None
        for el in versions['mindsdb']:
            if el['mindsdb_version'] is None:
                gui_lv = LooseVersion(el['gui_version'])
            else:
                mindsdb_lv = LooseVersion(el['mindsdb_version'])
                gui_lv = LooseVersion(el['gui_version'])
                if mindsdb_lv.vstring not in gui_versions or gui_lv > gui_versions[mindsdb_lv.vstring]:
                    gui_versions[mindsdb_lv.vstring] = gui_lv
                if max_mindsdb_lv is None or max_mindsdb_lv < mindsdb_lv:
                    max_mindsdb_lv = mindsdb_lv
            if max_gui_lv is None or max_gui_lv < gui_lv:
                max_gui_lv = gui_lv

        all_mindsdb_lv = [LooseVersion(x) for x in gui_versions.keys()]
        all_mindsdb_lv.sort()

        if current_mindsdb_lv.vstring in gui_versions:
            gui_version_lv = gui_versions[current_mindsdb_lv.vstring]
        elif current_mindsdb_lv > all_mindsdb_lv[-1]:
            gui_version_lv = max_gui_lv
        else:
            lower_versions = {key: value for key, value in gui_versions.items() if LooseVersion(key) < current_mindsdb_lv}
            if len(lower_versions) == 0:
                gui_version_lv = gui_versions[all_mindsdb_lv[0].vstring]
            else:
                all_lower_versions = [LooseVersion(x) for x in lower_versions.keys()]
                gui_version_lv = gui_versions[all_lower_versions[-1].vstring]
    except Exception as e:
        log.error(f'Error in compatible-config.json structure: {e}')
        return False

    current_gui_version = None

    version_txt_path = static_path.joinpath('version.txt')
    if version_txt_path.is_file():
        with open(version_txt_path, 'rt') as f:
            current_gui_version = f.readline()
    if current_gui_version is not None:
        current_gui_lv = LooseVersion(current_gui_version)
        if current_gui_lv >= gui_version_lv:
            return True

    log.info(f'New version of GUI available ({gui_version_lv.vstring}). Downloading...')

    shutil.rmtree(static_path)
    static_path.mkdir(parents=True, exist_ok=True)

    try:
        css_zip_path = str(static_path.joinpath('css.zip'))
        js_zip_path = str(static_path.joinpath('js.zip'))
        media_zip_path = str(static_path.joinpath('media.zip'))
        bucket = "https://mindsdb-web-builds.s3.amazonaws.com/"

        gui_version = gui_version_lv.vstring

        resources = [
            {
                'url': bucket + 'css-V' + gui_version + '.zip',
                'path': css_zip_path
            }, {
                'url': bucket + 'js-V' + gui_version + '.zip',
                'path': js_zip_path
            }, {
                'url': bucket + 'indexV' + gui_version + '.html',
                'path': str(static_path.joinpath('index.html'))
            }, {
                'url': bucket + 'favicon.ico',
                'path': str(static_path.joinpath('favicon.ico'))
            }, {
                'url': bucket + 'media.zip',
                'path': media_zip_path
            }
        ]

        def get_resources(resource):
            try:
                response = requests.get(resource['url'])
                if response.status_code != requests.status_codes.codes.ok:
                    return Exception(f"Error {response.status_code} GET {resource['url']}")
                open(resource['path'], 'wb').write(response.content)
            except Exception as e:
                return e
            return None

        for r in resources:
            get_resources(r)

        '''
        # to make downloading faster download each resource in a separate thread
        with concurrent.futures.ThreadPoolExecutor(max_workers=5) as executor:
            future_to_url = {executor.submit(get_resources, r): r for r in resources}
            for future in concurrent.futures.as_completed(future_to_url):
                res = future.result()
                if res is not None:
                    raise res
        '''

    except Exception as e:
        log.error(f'Error during downloading files from s3: {e}')
        return False

    static_folder = static_path.joinpath('static')
    static_folder.mkdir(parents=True, exist_ok=True)

    # unzip process
    for zip_path, dir_name in [[js_zip_path, 'js'], [css_zip_path, 'css']]:
        temp_dir = static_path.joinpath(f'temp_{dir_name}')
        temp_dir.mkdir(mode=0o777, exist_ok=True, parents=True)
        ZipFile(zip_path).extractall(temp_dir)
        files_path = static_path.joinpath('static', dir_name)
        if temp_dir.joinpath('build', 'static', dir_name).is_dir():
            shutil.move(temp_dir.joinpath('build', 'static', dir_name), files_path)
            shutil.rmtree(temp_dir)
        else:
            shutil.move(temp_dir, files_path)

    ZipFile(media_zip_path).extractall(static_folder)

    os.remove(js_zip_path)
    os.remove(css_zip_path)
    os.remove(media_zip_path)

    with open(version_txt_path, 'wt') as f:
        f.write(gui_version_lv.vstring)

    log.info(f'GUI version updated to {gui_version_lv.vstring}')
    return True


def initialize_flask(config, init_static_thread):
    # Apparently there's a bug that causes the static path not to work if it's '/' -- https://github.com/pallets/flask/issues/3134, I think '' should achieve the same thing (???)
    app = Flask(
        __name__,
        static_url_path='/static',
        static_folder=os.path.join(config.paths['static'], 'static/')
    )

    app.config['SEND_FILE_MAX_AGE_DEFAULT'] = 60
    app.config['SWAGGER_HOST'] = 'http://localhost:8000/mindsdb'
    authorizations = {
        'apikey': {
            'type': 'apiKey',
            'in': 'query',
            'name': 'apikey'
        }
    }

    api = Swagger_Api(
        app,
        authorizations=authorizations,
        security=['apikey'],
        url_prefix=':8000',
        prefix='/api',
        doc='/doc/'
    )

    port = config['api']['http']['port']
    host = config['api']['http']['host']

    # NOTE rewrite it, that hotfix to see GUI link
    log = logging.getLogger('mindsdb.http')
    url = f'http://{host}:{port}/'
    log.error(f' - GUI available at {url}')

    pid = os.getpid()
    x = threading.Thread(target=_open_webbrowser, args=(url, pid, port, init_static_thread, config.paths['static']), daemon=True)
    x.start()

    return app, api


def initialize_interfaces(config, app):
    app.default_store = DataStore(config)
    app.mindsdb_native = NativeInterface(config)
    app.custom_models = CustomModels(config)
    app.dbw = DatabaseWrapper(config)
    app.config_obj = config


def _open_webbrowser(url: str, pid: int, port: int, init_static_thread, static_folder):
    """Open webbrowser with url when http service is started.

    If some error then do nothing.
    """
    init_static_thread.join()
    inject_telemetry_to_static(static_folder)
    logger = logging.getLogger('mindsdb.http')
    try:
        is_http_active = wait_func_is_true(func=is_pid_listen_port, timeout=10,
                                           pid=pid, port=port)
        if is_http_active:
            webbrowser.open(url)
    except Exception as e:
        logger.error(f'Failed to open {url} in webbrowser with exception {e}')
        logger.error(traceback.format_exc())



# https://stackoverflow.com/questions/29298346/xmpp-sasl-scram-sha1-authentication
# https://tools.ietf.org/html/rfc5802

import base64
import hashlib
import hmac
import os

from pymongo.auth import _password_digest


class Scram():
    ''' implementation of server-side SCRAM-SHA-1 auth of mongodb
        TODO add SCRAM-SHA-256 auth
    '''

    def __init__(self, user, password):
        self.user = user
        self.password = password
        self.snonce = base64.b64encode(os.urandom(24)).decode()
        self.salt = base64.b64encode(os.urandom(16))
        self.iterations = 4096
        self.messages = []

    def process_client_first_message(self, payload):
        payload = payload[3:]
        payload_parts = self._split_payload(payload)
        self.client_user = payload_parts['n']
        self.unonce = payload_parts['r']
        self.messages.append(payload)

        responce_msg = f"r={self.unonce}{self.snonce},s={self.salt.decode()},i={self.iterations}"
        self.messages.append(responce_msg)
        return responce_msg

    def process_client_second_message(self, payload):
        self.messages.append(payload[:payload.rfind(',p=')])    # without 'p' part

        # TODO add user password check here

        messages = ','.join(self.messages)
        salted_password = self._salt_password()
        server_key = self._sha1_hmac(salted_password, b'Server Key')
        server_signature = self._sha1_hmac(server_key, messages.encode('utf-8'))
        return f'v={base64.b64encode(server_signature).decode()}'

    def _sha1_hmac(self, key, msg):
        return hmac.new(key, msg, digestmod=hashlib.sha1).digest()

    def _salt_password(self):
        password = _password_digest(self.user, self.password).encode("utf-8")
        return hashlib.pbkdf2_hmac(
            'sha1', password, base64.b64decode(self.salt), self.iterations
        )

    def _split_payload(self, payload):
        parts = {}
        for part in [x for x in payload.split(',') if len(x) > 0]:
            parts[part[0]] = part[2:]
        return parts



from mindsdb.api.mongo.classes import Responder
import mindsdb.api.mongo.functions as helpers


class Responce(Responder):
    when = {'collStats': helpers.is_true}

    def result(self, query, request_env, mindsdb_env, session):
        db = query['$db']
        collection = query['collStats']

        # NOTE real answer is huge, i removed most data from it.
        res = {
            'ns': "db.collection",
            'size': 1,
            'count': 0,
            'avgObjSize': 1,
            'storageSize': 16384,
            'capped': False,
            'wiredTiger': {
            },
            'nindexes': 1,
            'indexDetails': {
            },
            'totalIndexSize': 16384,
            'indexSizes': {
                '_id_': 16384
            },
            'ok': 1
        }

        res['ns'] = f"{db}.{collection}"
        if db == 'mindsdb' and collection == 'predictors':
            res['count'] = len(mindsdb_env['mindsdb_native'].get_models())

        return res


responder = Responce()



import datetime
import os
import threading
import tempfile
import re
import multipart

import mindsdb
from dateutil.parser import parse
from flask import request, send_file
from flask_restx import Resource, abort     # 'abort' using to return errors as json: {'message': 'error text'}
from flask import current_app as ca

from mindsdb.api.http.namespaces.configs.datasources import ns_conf
from mindsdb.api.http.namespaces.entitites.datasources.datasource import (
    datasource_metadata,
    put_datasource_params
)
from mindsdb.api.http.namespaces.entitites.datasources.datasource_data import (
    get_datasource_rows_params,
    datasource_rows_metadata
)
from mindsdb.api.http.namespaces.entitites.datasources.datasource_files import (
    put_datasource_file_params
)
from mindsdb.api.http.namespaces.entitites.datasources.datasource_missed_files import (
    datasource_missed_files_metadata,
    get_datasource_missed_files_params
)


def parse_filter(key, value):
    result = re.search(r'filter(_*.*)\[(.*)\]', key)
    operator = result.groups()[0].strip('_') or 'like'
    field = result.groups()[1]
    operators_map = {
        'like': 'like',
        'in': 'in',
        'nin': 'not in',
        'gt': '>',
        'lt': '<',
        'gte': '>=',
        'lte': '<=',
        'eq': '=',
        'neq': '!='
    }
    if operator not in operators_map:
        return None
    operator = operators_map[operator]
    return [field, operator, value]


@ns_conf.route('/')
class DatasourcesList(Resource):
    @ns_conf.doc('get_datasources_list')
    @ns_conf.marshal_list_with(datasource_metadata)
    def get(self):
        '''List all datasources'''
        return ca.default_store.get_datasources()


@ns_conf.route('/<name>')
@ns_conf.param('name', 'Datasource name')
class Datasource(Resource):
    @ns_conf.doc('get_datasource')
    @ns_conf.marshal_with(datasource_metadata)
    def get(self, name):
        '''return datasource metadata'''
        ds = ca.default_store.get_datasource(name)
        if ds is not None:
            return ds
        return '', 404

    @ns_conf.doc('delete_datasource')
    def delete(self, name):
        '''delete datasource'''
        try:
            ca.default_store.delete_datasource(name)
        except Exception as e:
            print(e)
            abort(400, str(e))
        return '', 200

    @ns_conf.doc('put_datasource', params=put_datasource_params)
    @ns_conf.marshal_with(datasource_metadata)
    def put(self, name):
        '''add new datasource'''
        data = {}

        def on_field(field):
            name = field.field_name.decode()
            value = field.value.decode()
            data[name] = value

        file_object = None

        def on_file(file):
            nonlocal file_object
            data['file'] = file.file_name.decode()
            file_object = file.file_object

        temp_dir_path = tempfile.mkdtemp(prefix='datasource_file_')

        if request.headers['Content-Type'].startswith('multipart/form-data'):
            parser = multipart.create_form_parser(
                headers=request.headers,
                on_field=on_field,
                on_file=on_file,
                config={
                    'UPLOAD_DIR': temp_dir_path.encode(),    # bytes required
                    'UPLOAD_KEEP_FILENAME': True,
                    'UPLOAD_KEEP_EXTENSIONS': True,
                    'MAX_MEMORY_FILE_SIZE': 0
                }
            )

            while True:
                chunk = request.stream.read(8192)
                if not chunk:
                    break
                parser.write(chunk)
            parser.finalize()
            parser.close()

            if file_object is not None and not file_object.closed:
                file_object.close()
        else:
            data = request.json

        if 'query' in data:
            source_type = request.json['integration_id']
            if source_type not in ca.default_store.config['integrations']:
                # integration doens't exist
                abort(400, f"{source_type} integration doesn't exist")

            if ca.default_store.config['integrations'][source_type]['type'] == 'mongodb':
                data['find'] = data['query']

            ca.default_store.save_datasource(name, source_type, data)
            os.rmdir(temp_dir_path)
            return ca.default_store.get_datasource(name)

        ds_name = data['name'] if 'name' in data else name
        source = data['source'] if 'source' in data else name
        source_type = data['source_type']

        if source_type == 'file':
            file_path = os.path.join(temp_dir_path, data['file'])
        else:
            file_path = None

        ca.default_store.save_datasource(ds_name, source_type, source, file_path)
        os.rmdir(temp_dir_path)

        return ca.default_store.get_datasource(ds_name)


ds_analysis = {}


def analyzing_thread(name, default_store):
    global ds_analysis
    ds_analysis[name] = None
    ds = default_store.get_datasource(name)
    analysis = default_store.get_analysis(ds['name'])
    ds_analysis[name] = {
        'created_at': datetime.datetime.utcnow(),
        'data': analysis
    }


@ns_conf.route('/<name>/analyze')
@ns_conf.param('name', 'Datasource name')
class Analyze(Resource):
    @ns_conf.doc('analyse_dataset')
    def get(self, name):
        global ds_analysis
        if name in ds_analysis:
            if ds_analysis[name] is None:
                return {'status': 'analyzing'}, 200
            else:
                analysis = ds_analysis[name]['data']
                return analysis, 200

        ds = ca.default_store.get_datasource(name)
        if ds is None:
            print('No valid datasource given')
            abort(400, 'No valid datasource given')

        x = threading.Thread(target=analyzing_thread, args=(name, ca.default_store))
        x.start()
        return {'status': 'analyzing'}, 200

@ns_conf.route('/<name>/analyze_refresh')
@ns_conf.param('name', 'Datasource name')
class Analyze(Resource):
    @ns_conf.doc('analyze_refresh_dataset')
    def get(self, name):
        global ds_analysis
        if name in ds_analysis:
            if ds_analysis[name] is None:
                return {'status': 'analyzing'}, 200
            else:
                del ds_analysis[name]

        ds = ca.default_store.get_datasource(name)
        if ds is None:
            print('No valid datasource given')
            abort(400, 'No valid datasource given')

        x = threading.Thread(target=analyzing_thread, args=(name, ca.default_store))
        x.start()
        return {'status': 'analyzing'}, 200


@ns_conf.route('/<name>/data/')
@ns_conf.param('name', 'Datasource name')
class DatasourceData(Resource):
    @ns_conf.doc('get_datasource_data', params=get_datasource_rows_params)
    @ns_conf.marshal_with(datasource_rows_metadata)
    def get(self, name):
        '''return data rows'''
        ds = ca.default_store.get_datasource(name)
        if ds is None:
            abort(400, 'No valid datasource given')

        params = {
            'page[size]': None,
            'page[offset]': None
        }
        where = []
        for key, value in request.args.items():
            if key == 'page[size]':
                params['page[size]'] = int(value)
            if key == 'page[offset]':
                params['page[offset]'] = int(value)
            elif key.startswith('filter'):
                param = parse_filter(key, value)
                if param is None:
                    abort(400, f'Not valid filter "{key}"')
                where.append(param)

        data_dict = ca.default_store.get_data(name, where, params['page[size]'], params['page[offset]'])

        return data_dict, 200


@ns_conf.route('/<name>/download')
@ns_conf.param('name', 'Datasource name')
class DatasourceMissedFilesDownload(Resource):
    @ns_conf.doc('get_datasource_download')
    def get(self, name):
        '''download uploaded file'''
        ds = ca.default_store.get_datasource(name)
        if not ds:
            abort(404, "{} not found".format(name))
        if not os.path.exists(ds['source']):
            abort(404, "{} not found".format(name))

        return send_file(os.path.abspath(ds['source']), as_attachment=True)



from mindsdb.integrations.clickhouse.clickhouse import Clickhouse
from mindsdb.integrations.postgres.postgres import PostgreSQL
from mindsdb.integrations.mariadb.mariadb import Mariadb
from mindsdb.integrations.mysql.mysql import MySQL
from mindsdb.integrations.mssql.mssql import MSSQL
from mindsdb.integrations.mongodb.mongodb import MongoDB

from mindsdb.utilities.log import log as logger


class DatabaseWrapper():

    def __init__(self, config):
        self.config = config

    def setup_integration(self, db_alias):
        try:
            # If this is the name of an integration
            integration = self._get_integration(db_alias)
            if integration != True:
                integration.setup()
        except Exception as e:
            logger.warning('Failed to integrate with database ' + integration.name + f', error: {e}')

    def _get_integration(self, db_alias):
        if self.config['integrations'][db_alias]['publish']:
            db_type = self.config['integrations'][db_alias]['type']
            if db_type == 'clickhouse':
                return Clickhouse(self.config, db_alias)
            elif db_type == 'mariadb':
                return Mariadb(self.config, db_alias)
            elif db_type == 'mysql':
                return MySQL(self.config, db_alias)
            elif db_type == 'postgres':
                return PostgreSQL(self.config, db_alias)
            elif db_type == 'mssql':
                return MSSQL(self.config, db_alias)
            elif db_type == 'mongodb':
                return MongoDB(self.config, db_alias)
            else:
                logger.warning(f'Uknown integration type: {db_type} for database called: {db_alias}')
            return False
        return True

    def _get_integrations(self):
        integrations = [self._get_integration(x) for x in self.config['integrations']]
        integrations = [x for x in integrations if x != True and x != False]
        return integrations

    def register_predictors(self, model_data_arr):
        for integration in self._get_integrations():
            if integration.check_connection():
                integration.register_predictors(model_data_arr)
            else:
                logger.warning(f"There is no connection to {integration.name}. predictor wouldn't be registred.")

    def unregister_predictor(self, name):
        for integration in self._get_integrations():
            if integration.check_connection():
                integration.unregister_predictor(name)
            else:
                logger.warning(f"There is no connection to {integration.name}. predictor wouldn't be unregistred")

    def check_connections(self):
        connections = {}
        for integration in self._get_integrations():
            connections[integration.name] = integration.check_connection()

        return connections



import copy
import os
import shutil
import time
from io import BytesIO

from dateutil.parser import parse as parse_datetime
from flask import request, send_file
from flask_restx import Resource, abort
from flask import current_app as ca

from mindsdb.api.http.namespaces.configs.predictors import ns_conf
from mindsdb.api.http.namespaces.entitites.predictor_metadata import (
    predictor_metadata,
    predictor_query_params,
    upload_predictor_params,
    put_predictor_params
)
from mindsdb.api.http.namespaces.entitites.predictor_status import predictor_status


def debug_pkey_type(model, keys=None, reset_keyes=True, type_to_check=list, append_key=True):
    if type(model) != dict:
        return
    for k in model:
        if reset_keyes:
            keys = []
        if type(model[k]) == dict:
            keys.append(k)
            debug_pkey_type(model[k], copy.deepcopy(keys), reset_keyes=False)
        if type(model[k]) == type_to_check:
            print(f'They key {keys}->{k} has type list')
        if type(model[k]) == list:
            for item in model[k]:
                debug_pkey_type(item, copy.deepcopy(keys), reset_keyes=False)


def preparse_results(results, format_flag='explain'):
    response_arr = []
    for res in results:
        if format_flag == 'explain':
            response_arr.append(res.explain())
        elif format_flag == 'epitomize':
            response_arr.append(res.epitomize())
        elif format_flag == 'new_explain':
            response_arr.append(res.explanation)
        else:
            response_arr.append(res.explain())

    if len(response_arr) > 0:
        return response_arr
    else:
        abort(400, "")

def is_custom(name):
    if name in [x['name'] for x in ca.custom_models.get_models()]:
        return True
    return False

@ns_conf.route('/')
class PredictorList(Resource):
    @ns_conf.doc('list_predictors')
    @ns_conf.marshal_list_with(predictor_status, skip_none=True)
    def get(self):
        '''List all predictors'''

        return [*ca.mindsdb_native.get_models(),*ca.custom_models.get_models()]

@ns_conf.route('/custom/<name>')
@ns_conf.param('name', 'The predictor identifier')
@ns_conf.response(404, 'predictor not found')
class CustomPredictor(Resource):
    @ns_conf.doc('put_custom_predictor')
    def put(self, name):
        try:
            trained_status = request.json['trained_status']
        except:
            trained_status = 'untrained'

        predictor_file = request.files['file']
        fpath = os.path.join(ca.config_obj.paths['tmp'],  name + '.zip')
        with open(fpath, 'wb') as f:
            f.write(predictor_file.read())

        ca.custom_models.load_model(fpath, name, trained_status)

        return f'Uploaded custom model {name}'

@ns_conf.route('/<name>')
@ns_conf.param('name', 'The predictor identifier')
@ns_conf.response(404, 'predictor not found')
class Predictor(Resource):
    @ns_conf.doc('get_predictor')
    @ns_conf.marshal_with(predictor_metadata, skip_none=True)
    def get(self, name):
        try:
            if is_custom(name):
                model = ca.custom_models.get_model_data(name)
            else:
                model = ca.mindsdb_native.get_model_data(name, db_fix=False)
        except Exception as e:
            abort(404, "")

        for k in ['train_end_at', 'updated_at', 'created_at']:
            if k in model and model[k] is not None:
                model[k] = parse_datetime(model[k])

        return model

    @ns_conf.doc('delete_predictor')
    def delete(self, name):
        '''Remove predictor'''
        if is_custom(name):
            ca.custom_models.delete_model(name)
        else:
            ca.mindsdb_native.delete_model(name)

        return '', 200

    @ns_conf.doc('put_predictor', params=put_predictor_params)
    def put(self, name):
        '''Learning new predictor'''
        data = request.json
        to_predict = data.get('to_predict')

        try:
            kwargs = data.get('kwargs')
        except:
            kwargs = None

        if type(kwargs) != type({}):
            kwargs = {}

        if 'equal_accuracy_for_all_output_categories' not in kwargs:
            kwargs['equal_accuracy_for_all_output_categories'] = True

        if 'advanced_args' not in kwargs:
            kwargs['advanced_args'] = {}

        if 'use_selfaware_model' not in kwargs['advanced_args']:
            kwargs['advanced_args']['use_selfaware_model'] = False

        try:
            retrain = data.get('retrain')
            if retrain in ('true', 'True'):
                retrain = True
            else:
                retrain = False
        except Exception:
            retrain = None

        ds_name = data.get('data_source_name') if data.get('data_source_name') is not None else data.get('from_data')
        from_data = ca.default_store.get_datasource_obj(ds_name, raw=True)

        if from_data is None:
            return {'message': f'Can not find datasource: {ds_name}'}, 400

        if retrain is True:
            original_name = name
            name = name + '_retrained'

        ca.mindsdb_native.learn(name, from_data, to_predict, kwargs)
        for i in range(20):
            try:
                # Dirty hack, we should use a messaging queue between the predictor process and this bit of the code
                ca.mindsdb_native.get_model_data(name)
                break
            except Exception:
                time.sleep(1)

        if retrain is True:
            try:
                ca.mindsdb_native.delete_model(original_name)
                ca.mindsdb_native.rename_model(name, original_name)
            except:
                pass

        return '', 200

@ns_conf.route('/<name>/learn')
@ns_conf.param('name', 'The predictor identifier')
class PredictorLearn(Resource):
    def post(self, name):
        data = request.json
        to_predict = data.get('to_predict')
        kwargs = data.get('kwargs', None)

        if type(kwargs) != type({}):
            kwargs = {}

        if 'advanced_args' not in kwargs:
            kwargs['advanced_args'] = {}


        ds_name = data.get('data_source_name') if data.get('data_source_name') is not None else data.get('from_data')
        from_data = ca.default_store.get_datasource_obj(ds_name, raw=True)

        ca.custom_models.learn(name, from_data, to_predict, kwargs)

        return '', 200


@ns_conf.route('/<name>/predict')
@ns_conf.param('name', 'The predictor identifier')
class PredictorPredict(Resource):
    @ns_conf.doc('post_predictor_predict', params=predictor_query_params)
    def post(self, name):
        '''Queries predictor'''

        data = request.json

        when = data.get('when') or {}
        try:
            format_flag = data.get('format_flag')
        except:
            format_flag = 'explain'

        try:
            kwargs = data.get('kwargs')
        except:
            kwargs = {}

        if type(kwargs) != type({}):
            kwargs = {}

        if is_custom(name):
            return ca.custom_models.predict(name, when_data=when, **kwargs)
        else:
            results = ca.mindsdb_native.predict(name, when_data=when, **kwargs)

        return preparse_results(results, format_flag)


@ns_conf.route('/<name>/predict_datasource')
@ns_conf.param('name', 'The predictor identifier')
class PredictorPredictFromDataSource(Resource):
    @ns_conf.doc('post_predictor_predict', params=predictor_query_params)
    def post(self, name):
        data = request.json

        from_data = ca.default_store.get_datasource_obj(data.get('data_source_name'))
        if from_data is None:
            abort(400, 'No valid datasource given')

        try:
            format_flag = data.get('format_flag')
        except Exception:
            format_flag = 'explain'

        try:
            kwargs = data.get('kwargs')
        except Exception:
            kwargs = {}

        if not isinstance(kwargs, dict):
            kwargs = {}

        if is_custom(name):
            return ca.custom_models.predict(name, from_data=from_data, **kwargs)

        results = ca.mindsdb_native.predict(name, when_data=from_data, **kwargs)
        return preparse_results(results, format_flag)


@ns_conf.route('/upload')
class PredictorUpload(Resource):
    @ns_conf.doc('predictor_query', params=upload_predictor_params)
    def post(self):
        '''Upload existing predictor'''
        predictor_file = request.files['file']
        # @TODO: Figure out how to remove
        fpath = os.path.join(ca.config_obj.paths['tmp'], 'new.zip')
        with open(fpath, 'wb') as f:
            f.write(predictor_file.read())

        ca.mindsdb_native.load_model(fpath)
        try:
            os.remove(fpath)
        except Exception:
            pass

        return '', 200


@ns_conf.route('/<name>/download')
@ns_conf.param('name', 'The predictor identifier')
class PredictorDownload(Resource):
    @ns_conf.doc('get_predictor_download')
    def get(self, name):
        '''Export predictor to file'''
        ca.mindsdb_native.export_model(name)
        fname = name + '.zip'
        original_file = os.path.join(fname)
        # @TODO: Figure out how to remove
        fpath = os.path.join(ca.config_obj.paths['tmp'], fname)
        shutil.move(original_file, fpath)

        with open(fpath, 'rb') as f:
            data = BytesIO(f.read())

        try:
            os.remove(fpath)
        except Exception as e:
            pass

        return send_file(
            data,
            mimetype='application/zip',
            attachment_filename=fname,
            as_attachment=True
        )

@ns_conf.route('/<name>/rename')
@ns_conf.param('name', 'The predictor identifier')
class PredictorDownload(Resource):
    @ns_conf.doc('get_predictor_download')
    def get(self, name):
        '''Export predictor to file'''
        try:
            new_name = request.args.get('new_name')
            if is_custom(name):
                ca.custom_models.rename_model(name, new_name)
            else:
                ca.mindsdb_native.rename_model(name, new_name)
        except Exception as e:
            return str(e), 400

        return f'Renamed model to {new_name}', 200



class ModelInterface:
    # Should not be override, or if it is, should be addtionally called using `super(YourChildClass, self).initialize_column_types()`
    def initialize_column_types(self):
        if self.column_type_map is None:
            self.column_type_map = {
                'Empty_target': {
                    'typing': {
                        'data_type': 'Text'
                        ,'data_subtype': 'Short Text'
                    }
                }
                ,'Empty_input': {
                    'typing': {
                        'data_type': 'Text'
                        ,'data_subtype': 'Short Text'
                    }
                }
            }

        if self.to_predict is None:
            self.to_predict = 'Empty_target'

    # Optional methods
    def setup(self):
        pass

    # Optional but with a default implementation
    column_type_map = None
    to_predict = None

    def save(self, path):
        import pickle
        with open(path, 'wb') as fp:
            pickle.dump(self, fp)

    @staticmethod
    def load(path):
        import pickle
        with open(path, 'rb') as fp:
            model = pickle.load(fp)
        return model

    # Mandatory under certain circumstances
    def fit(self, from_data, to_predict, data_analysis, kwargs):
        self.to_predict = to_predict
        self.column_type_map = data_analysis
        raise NotImplementedError('You must implement `fit` in order to be able to train your model')

    # Mandatory
    def predict(self, from_data, kwargs):
        raise NotImplementedError('You must implement `predict` in order to be able to use your model')



import pymssql
from mindsdb.integrations.base import Integration


class MSSQL(Integration):
    def _get_connnection(self):
        integration = self.config['integrations'][self.name]
        return pymssql.connect(
            server=integration['host'],
            host=integration['host'],
            user=integration['user'],
            password=integration['password'],
            database=integration.get('database', 'master'),
            port=integration['port'],
            autocommit=True  # that need for CRUD operations
        )

    def _query(self, query, fetch=False):
        conn = self._get_connnection()
        cur = conn.cursor(as_dict=True)
        cur.execute(query)
        res = True
        if fetch:
            res = cur.fetchall()
        cur.close()
        conn.close()
        return res

    def setup(self):
        integration = self.config['integrations'][self.name]
        driver_name = integration.get('odbc_driver_name', 'MySQL ODBC 8.0 Unicode Driver')
        servers = self._query('exec sp_linkedservers;', fetch=True)
        servers = [x['SRV_NAME'] for x in servers]
        if self.mindsdb_database in servers:
            self._query(f"exec sp_dropserver @server = N'{self.mindsdb_database}';")
        mysql = self.config['api']['mysql']
        self._query(f'''
            exec sp_addlinkedserver
                @server = N'{self.mindsdb_database}'
                ,@srvproduct=N'MySQL'
                ,@provider=N'MSDASQL'
                ,@provstr=N'DRIVER={{{driver_name}}}; SERVER={mysql['host']}; PORT={mysql['port']}; DATABASE=mindsdb; USER={mysql['user']}_{self.name}; {('PASSWORD=' + mysql['password'] + ';') if len(mysql['password']) > 0 else ''} OPTION=3;';
        ''')
        try:
            self._query(f"exec sp_serveroption @server='{self.mindsdb_database}', @optname='rpc', @optvalue='true'")
            self._query(f"exec sp_serveroption @server='{self.mindsdb_database}', @optname='rpc out', @optvalue='true'")
        except Exception:
            # nothing critical if server options not setted. Only 'four part' notation will not work.
            print('MSSQL integration: failed to set server options.')

    def register_predictors(self, model_data_arr):
        pass

    def unregister_predictor(self, name):
        pass

    def check_connection(self):
        try:
            conn = self._get_connnection()
            conn.close()
            connected = True
        except Exception:
            connected = False
        return connected



from mindsdb.api.http.namespaces.configs.predictors import ns_conf
from mindsdb.api.http.namespaces.entitites.column_metadata import column_metadata

from flask_restx import fields

nested_histogram_data = ns_conf.model('NestedHistogramData', {
    'x': fields.List(fields.String, required=False, description='Ordered labels'),
    #'y': fields.List(fields.Float, required=False, description='Count for each label'),
    'y': fields.List(fields.Raw, required=False, description='Count for each label'),
    'x_explained': fields.List(fields.List( fields.Nested(column_metadata)), required=False, description='Ordered list of lists where each element in the histogram has a list of column metadata only relevant to each  subset of data defined by the histogram bucket '),

})



from flask_restx import fields
from collections import OrderedDict

from mindsdb.api.http.namespaces.configs.datasources import ns_conf

get_datasource_missed_files_params = OrderedDict([
    ('page[size]', {
        'description': 'page size',
        'type': 'integer',
        'in': 'path',
        'required': False,
        'default': 0
    }),
    ('page[number]', {
        'description': 'start page',
        'type': 'integer',
        'in': 'path',
        'required': False,
        'default': 0
    })
])

datasource_missed_file_metadata = ns_conf.model('DatasourceMissedFile', {
    'column_name': fields.String(required=False, description='file column name'),
    'index': fields.Integer(required=False, description='row index in datasource'),
    'path': fields.String(required=False, description='file column value')
})

datasource_missed_files_metadata = ns_conf.model('DatasourceMissedFiles', {
    'rowcount': fields.Integer(required=False, description='number of missed files'),
    'data': fields.List(fields.Nested(datasource_missed_file_metadata), required=False, description='columns description')
})

EXAMPLES = [{
    'rowcount': 2,
    'data': [{
        'column_name': 'file',
        'index': 3,
        'path': '/tmp/3.jpg'
    }, {
        'column_name': 'file',
        'index': 4,
        'path': '/tmp/4.jpg'
    }]
}]



from .scram import Scram


class Session():
    def __init__(self, config):
        self.config = config

    def init_scram(self):
        user = self.config['api']['mongodb'].get('user', '')
        password = self.config['api']['mongodb'].get('password', '')
        self.scram = Scram(user, password)



from mindsdb.api.http.namespaces.configs.predictors import ns_conf

from mindsdb.api.http.namespaces.entitites.histogram_data import histogram_data
from mindsdb.api.http.namespaces.entitites.label_group import label_group

from flask_restx import fields

data_distribution_metadata = ns_conf.model('DataDistributionMetadata', {
    'data_histogram': fields.Nested(histogram_data, required=False, description='The histogram representing the data in this column if possible'),
    'clusters': fields.List(fields.Nested(label_group), required=False, description='The labels per cluster'),
    'mean': fields.String(required=False, description='The mean value if possible, encoded as string')
})


from mindsdb.api.http.namespaces.configs.predictors import ns_conf
from flask_restx import fields


quality_metric = ns_conf.model('QualityMetric', {
    'type': fields.String(required=False, description='The quality type', enum=['error', 'warning', 'info']),
    'score': fields.Float(required=False, description='The score on the specific metric value 0-1'),
    'description': fields.String(required=False, description='The quality metric description'),
    'warning': fields.String(required=False, description=''),
    'name': fields.String(required=False, description=''),
})



class DataNode:
    type = 'meta'

    def __init__(self):
        pass

    def getType(self):
        return self.type

    def getTables(self):
        pass

    def hasTable(self, tableName):
        pass

    def getTableColumns(self, tableName):
        pass

    def select(self, table=None, columns=None, where=None, where_data=None, order_by=None, group_by=None, came_from=None):
        return []



from mindsdb.api.mongo.classes import Responder
import mindsdb.api.mongo.functions as helpers


class Responce(Responder):
    when = {'listDatabases': helpers.is_true}

    def result(self, query, request_env, mindsdb_env, session):
        db = mindsdb_env['config']['api']['mongodb']['database']

        return {
            'databases': [{
                'name': db,
                'sizeOnDisk': 1 << 16,
                'empty': False
            }, {
                'name': 'admin',
                'sizeOnDisk': 1 << 16,
                'empty': False
            }],
            'ok': 1,
        }


responder = Responce()



import os
import sys
import logging


class LoggerWrapper(object):
    def __init__(self, writer):
        self._writer = writer
        self._msg = ''

    def write(self, message):
        self._msg = self._msg + message
        while '\n' in self._msg:
            pos = self._msg.find('\n')
            self._writer(self._msg[:pos])
            self._msg = self._msg[pos + 1:]

    def flush(self):
        if self._msg != '':
            self._writer(self._msg)
            self._msg = ''


def initialize_log(config, logger_name='main', wrap_print=False):
    ''' Create new logger
    :param config: object, app config
    :param logger_name: str, name of logger
    :param wrap_print: bool, if true, then print() calls will be wrapped by log.debug() function.
    '''
    log = logging.getLogger(f'mindsdb.{logger_name}')
    log.propagate = False
    log.setLevel(min(
        getattr(logging, config['log']['level']['console']),
        getattr(logging, config['log']['level']['file'])
    ))

    formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')

    ch = logging.StreamHandler()
    ch.setLevel(config['log']['level']['console'])       # that level will be in console
    log.addHandler(ch)

    log_path = os.path.join(config.paths['log'], logger_name)
    if not os.path.isdir(log_path):
        os.mkdir(log_path)

    fh = logging.handlers.RotatingFileHandler(
        os.path.join(log_path, 'log.txt'),
        mode='a',
        encoding='utf-8',
        maxBytes=100 * 1024,
        backupCount=3
    )
    fh.setLevel(config['log']['level']['file'])
    fh.setFormatter(formatter)
    log.addHandler(fh)

    if wrap_print:
        sys.stdout = LoggerWrapper(log.debug)


log = logging.getLogger('mindsdb')



from mindsdb.api.mysql.mysql_proxy.datahub.datanodes.datanode import DataNode
from mindsdb.interfaces.datastore.datastore import DataStore


class DataSourceDataNode(DataNode):
    type = 'mindsdb-datasource'

    def __init__(self, config):
        self.config = config
        self.datastore = DataStore(config)
        # self.mindsdb_native = NativeInterface(config)

    def getTables(self):
        dss = self.datastore.get_datasources()
        return [x['name'] for x in dss]

    def hasTable(self, table):
        return table in self.getTables()

    def getTableColumns(self, table):
        ds = self.datastore.get_datasource(table)
        return [x['name'] for x in ds['columns']]

    def select(self, table, columns=None, where=None, where_data=None, order_by=None, group_by=None, came_from=None):
        data = self.datastore.get_data(table, where=None, limit=None, offset=None)
        return data['data']



import socketserver as SocketServer
import socket
import struct
import bson
from bson import codec_options
from collections import OrderedDict
from abc import abstractmethod

from mindsdb.api.mongo.classes import RespondersCollection, Session

from mindsdb.api.mongo.responders import responders
import mindsdb.api.mongo.functions as helpers
from mindsdb.api.mongo.utilities import log

from mindsdb.interfaces.datastore.datastore import DataStore
from mindsdb.interfaces.native.native import NativeInterface

OP_REPLY = 1
OP_UPDATE = 2001
OP_INSERT = 2002
OP_QUERY = 2004
OP_GET_MORE = 2005
OP_DELETE = 2006
OP_KILL_CURSORS = 2007
OP_MSG = 2013

BYTE = '<b'
INT = '<i'
UINT = '<I'
LONG = '<q'


def unpack(format, buffer, start=0):
    end = start + struct.calcsize(format)
    return struct.unpack(format, buffer[start:end])[0], end


def get_utf8_string(buffer, start=0):
    end = buffer.index(b"\x00", start)
    s = buffer[start:end].decode('utf8')
    return s, end + 1


CODEC_OPTIONS = codec_options.CodecOptions(document_class=OrderedDict)


def decode_documents(buffer, start, content_size):
    docs = bson.decode_all(buffer[start:start + content_size], CODEC_OPTIONS)
    return docs, start + content_size


class OperationResponder():
    def __init__(self, responders):
        self.responders = responders

    @abstractmethod
    def handle(self, query_bytes):
        pass

    @abstractmethod
    def to_bytes(self, response, request_id):
        pass

    def get_match_responder(self, query):
        for responder in self.responders:
            if responder['match'](query):
                return responder['response']
        raise Exception('is no responder')


# NOTE probably, it need only for mongo version < 3.6
class OpInsertResponder(OperationResponder):
    def handle(self, buffer, request_id, mindsdb_env, session):
        flags, pos = unpack(UINT, buffer)
        namespace, pos = get_utf8_string(buffer, pos)
        query = bson.decode_all(buffer[pos:], CODEC_OPTIONS)
        responder = self.responders.find_match(query)
        assert responder is not None, 'query cant be processed'

        request_args = {
            'request_id': request_id
        }

        documents = responder.handle(query, request_args, mindsdb_env, session)

        return documents

    def to_bytes(self, response, request_id):
        pass


OP_MSG_FLAGS = {
    'checksumPresent': 0,
    'moreToCome': 1,
    'exhaustAllowed': 16
}


# NOTE used in mongo version > 3.6
class OpMsgResponder(OperationResponder):
    def handle(self, buffer, request_id, mindsdb_env, session):
        query = OrderedDict()
        flags, pos = unpack(UINT, buffer)

        checksum_present = bool(flags & (1 << OP_MSG_FLAGS['checksumPresent']))
        if checksum_present:
            msg_len = len(buffer) - 4
        else:
            msg_len = len(buffer)

        # sections
        while pos < msg_len:
            kind, pos = unpack(BYTE, buffer, pos)
            if kind == 0:
                # body
                section_size, _ = unpack(INT, buffer, pos)
                docs, pos = decode_documents(buffer, pos, section_size)
                query.update(docs[0])
            elif kind == 1:
                # Document
                section_size, pos = unpack(INT, buffer, pos)
                seq_id, pos = get_utf8_string(buffer, pos)
                docs_len = section_size - struct.calcsize(INT) - len(seq_id) - 1
                docs, pos = decode_documents(buffer, pos, docs_len)
                query[seq_id] = docs

        remaining = len(buffer) - pos
        if checksum_present:
            if remaining != 4:
                raise Exception('should be checksum at the end of message')
            # TODO read and check checksum
        elif remaining != 0:
            raise Exception('is bytes left after msg parsing')

        log.debug(f'GET OpMSG={query}')

        responder = self.responders.find_match(query)
        assert responder is not None, 'query cant be processed'

        request_args = {
            'request_id': request_id,
            'database': query['$db']
        }

        documents = responder.handle(query, request_args, mindsdb_env, session)

        return documents

    def to_bytes(self, response, request_id):
        flags = struct.pack("<I", 0)  # TODO
        payload_type = struct.pack("<b", 0)  # TODO
        payload_data = bson.BSON.encode(response)
        data = b''.join([flags, payload_type, payload_data])

        reply_id = 0  # TODO add seq here
        response_to = request_id

        header = struct.pack("<iiii", 16 + len(data), reply_id, response_to, OP_MSG)
        return header + data


# NOTE used in any mongo shell version
class OpQueryResponder(OperationResponder):
    def handle(self, buffer, request_id, mindsdb_env, session):
        # https://docs.mongodb.com/manual/reference/mongodb-wire-protocol/#wire-op-query
        flags, pos = unpack(UINT, buffer)
        namespace, pos = get_utf8_string(buffer, pos)
        is_command = namespace.endswith('.$cmd')
        num_to_skip, pos = unpack(INT, buffer, pos)
        num_to_return, pos = unpack(INT, buffer, pos)
        docs = bson.decode_all(buffer[pos:], CODEC_OPTIONS)

        query = docs[0]  # docs = [query, returnFieldsSelector]

        log.debug(f'GET OpQuery={query}')

        responder = self.responders.find_match(query)
        assert responder is not None, 'query cant be processed'

        request_args = {
            'num_to_skip': num_to_skip,
            'num_to_return': num_to_return,
            'request_id': request_id,
            'is_command': is_command
        }

        documents = responder.handle(query, request_args, mindsdb_env, session)

        return documents

    def to_bytes(self, request, request_id):
        flags = struct.pack("<i", 0)  # TODO
        cursor_id = struct.pack("<q", 0)  # TODO
        starting_from = struct.pack("<i", 0)  # TODO
        number_returned = struct.pack("<i", len([request]))
        reply_id = 123  # TODO
        response_to = request_id

        log.debug(f'RET docs={request}')

        data = b''.join([flags, cursor_id, starting_from, number_returned])
        data += b''.join([bson.BSON.encode(doc) for doc in [request]])

        message = struct.pack("<i", 16 + len(data))
        message += struct.pack("<i", reply_id)
        message += struct.pack("<i", response_to)
        message += struct.pack("<i", OP_REPLY)

        return message + data


class MongoRequestHandler(SocketServer.BaseRequestHandler):
    _stopped = False

    def _init_ssl(self):
        import ssl
        import tempfile
        import atexit
        import os

        from mindsdb.utilities.wizards import make_ssl_cert

        CERT_PATH = tempfile.mkstemp(prefix='mindsdb_cert_', text=True)[1]
        make_ssl_cert(CERT_PATH)
        atexit.register(lambda: os.remove(CERT_PATH))

        ssl_context = ssl.SSLContext()
        ssl_context.load_cert_chain(CERT_PATH)
        ssl_socket = ssl_context.wrap_socket(
            self.request,
            server_side=True,
            do_handshake_on_connect=True
        )
        self.request = ssl_socket

    def handle(self):
        log.debug('connect')
        log.debug(str(self.server.socket))

        self.session = Session(self.server.mindsdb_env['config'])

        first_byte = self.request.recv(1, socket.MSG_PEEK)
        if first_byte == b'\x16':
            # TLS 'client hello' starts from \x16
            self._init_ssl()

        while True:
            header = self._read_bytes(16)
            if header is False:
                # connection closed by client
                break
            length, pos = unpack(INT, header)
            request_id, pos = unpack(INT, header, pos)
            response_to, pos = unpack(INT, header, pos)
            opcode, pos = unpack(INT, header, pos)
            log.debug(f'GET length={length} id={request_id} opcode={opcode}')
            msg_bytes = self._read_bytes(length - pos)
            answer = self.get_answer(request_id, opcode, msg_bytes)
            self.request.send(answer)

    def get_answer(self, request_id, opcode, msg_bytes):
        if opcode not in self.server.operationsHandlersMap:
            raise NotImplementedError(f'Unknown opcode {opcode}')
        responder = self.server.operationsHandlersMap[opcode]
        assert responder is not None, 'error'
        response = responder.handle(msg_bytes, request_id, self.server.mindsdb_env, self.session)
        assert response is not None, 'error'
        return responder.to_bytes(response, request_id)

    def _read_bytes(self, length):
        buffer = b''
        while length:
            chunk = self.request.recv(length)
            if chunk == b'':
                log.debug('Connection closed')
                return False

            length -= len(chunk)
            buffer += chunk
        return buffer


class MongoServer(SocketServer.ThreadingMixIn, SocketServer.TCPServer):
    def __init__(self, config):
        mongodb_config = config['api'].get('mongodb')
        assert mongodb_config is not None, 'is no mongodb config!'
        host = mongodb_config['host']
        port = mongodb_config['port']
        log.debug(f'start mongo server on {host}:{port}')

        super().__init__((host, int(port)), MongoRequestHandler)

        self.mindsdb_env = {
            'config': config,
            'data_store': DataStore(config),
            'mindsdb_native': NativeInterface(config)
        }

        respondersCollection = RespondersCollection()

        opQueryResponder = OpQueryResponder(respondersCollection)
        opMsgResponder = OpMsgResponder(respondersCollection)
        opInsertResponder = OpInsertResponder(respondersCollection)

        self.operationsHandlersMap = {
            OP_QUERY: opQueryResponder,
            OP_MSG: opMsgResponder,
            OP_INSERT: opInsertResponder
        }

        respondersCollection.add(
            when={'drop': 'system.sessions'},
            result={'ok': 1}
        )
        respondersCollection.add(
            when={'update': 'system.version'},
            result={'ok': 1}
        )
        respondersCollection.add(
            when={'setFeatureCompatibilityVersion': helpers.is_true},
            result={'ok': 1}
        )
        # OpMSG=OrderedDict([('features', 1), ('$clusterTime', OrderedDict([('clusterTime', Timestamp(1599748325, 1)), ('signature', OrderedDict([('hash', b'\xb8\xc3\x03\x18\xca\xe6bh\xf0\xcb47,\x924\x8a >\xfc\x91'), ('keyId', 6870854312365391875)]))])), ('$configServerState', OrderedDict([('opTime', OrderedDict([('ts', Timestamp(1599748325, 1)), ('t', 1)]))])), ('$db', 'admin')])
        respondersCollection.add(
            when={'features': helpers.is_true},
            result={'ok': 1}
        )
        # OpMSG=OrderedDict([('serverStatus', 1), ('$clusterTime', OrderedDict([('clusterTime', Timestamp(1599748366, 1)), ('signature', OrderedDict([('hash', b'\xa1E}\xbbIU\xc2D\x95++\x82\x88\xb5\x84\xf5\xda)+B'), ('keyId', 6870854312365391875)]))])), ('$configServerState', OrderedDict([('opTime', OrderedDict([('ts', Timestamp(1599748366, 1)), ('t', 1)]))])), ('$db', 'admin')])
        respondersCollection.add(
            when={'serverStatus': helpers.is_true},
            result={'ok': 1}
        )
        # OpMSG=OrderedDict([('ismaster', 1), ('$db', 'admin'), ('$clusterTime', OrderedDict([('clusterTime', Timestamp(1599749031, 1)), ('signature', OrderedDict([('hash', b'6\x87\xd5Y\xa7\xc7\xcf$\xab\x1e\xa2{\xe5B\xe5\x99\xdbl\x8d\xf4'), ('keyId', 6870854312365391875)]))])), ('$client', OrderedDict([('application', OrderedDict([('name', 'MongoDB Shell')])), ('driver', OrderedDict([('name', 'MongoDB Internal Client'), ('version', '3.6.3')])), ('os', OrderedDict([('type', 'Linux'), ('name', 'Ubuntu'), ('architecture', 'x86_64'), ('version', '18.04')])), ('mongos', OrderedDict([('host', 'maxs-comp:27103'), ('client', '127.0.0.1:52148'), ('version', '3.6.3')]))])), ('$configServerState', OrderedDict([('opTime', OrderedDict([('ts', Timestamp(1599749031, 1)), ('t', 1)]))]))])

        respondersCollection.responders += responders


def run_server(config):
    SocketServer.TCPServer.allow_reuse_address = True
    with MongoServer(config) as srv:
        srv.serve_forever()
